# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""Main file for performing operations on entity manager app under Activate.

'Classifiers', 'Classifier', 'Tags' , 'TagSet', 'Tag', 'EntityManagerTypes' , `ActivateEntities`, and `ActivateEntity` are 8 classes defined in this file.

ActivateEntities:   Class for representing all the regex entities in the commcell.

ActivateEntity:     Class for representing a single regex entity in the commcell.

EntityManagerTypes: Class to represent different entity types in entity manager

Tags:   Class to represent TagSets in the commcell

TagSet: Class to represent single Tagset entity in the commcell

Tag:  Class to represent tag inside a TagSet

Classifiers:    Class to represent Classifiers entities in the commcell

Classifier:     Class to represent a single classifier entity in the commcell

Tags:

    __init__()                          --  initialise object of the Tags class

     _response_not_success()            --  parses through the exception response, and raises SDKException

    refresh()                           --  refresh the TagSet associated with the commcell

    _get_tag_sets_from_collections()    --  gets all the TagSet details from collection response

    _get_all_tag_sets()                 --  Returns dict consisting all TagSets associated with commcell

    get_properties()                    --  Returns the properties for the given TagSet name

    has_tag_set()                       --  Checks whether tagset with given name exists in commcell or not

    add()                               --  Creates new TagSet in the commcell

    delete()                            --  Deletes the Tagset in the commcell

    get()                               --  Returns the TagSet object for given tagset name


TagSet:

     __init__()                         --  initialize an object of TagSet Class with the given tagset
                                                name and id

     _response_not_success()            --  parses through the exception response, and raises SDKException

    refresh()                           --  refresh the properties of the TagSet

    modify()                            --  Modifies the tagset in the commcell

    share()                             --  Shares tagset with user or group in the commcell

    add_tag()                           --  Creates new tag inside this tagset container in commcell

    get()                               --  Returns the Tag class object for given tag name

    has_tag()                           --  checks whether given tag exists in tagset or not

    get_tag_id()                        --  Returns the tag id for given tag name

    _get_tag_set_id()                   --  Gets tag set container id for the given Tagset name

    _get_tag_set_properties()           --  Gets all the details of associated Tagset

TagSet Attributes
-----------------

    **guid**            --  returns container GUID of this tagset

    **full_name**       --  returns the full name of tagset container

    **comment**         --  returns the comment for this tagset

    **owner**           --  returns the owner user name for this tagset

    **owner_alias_name** -  returns the owner alias name for this tagset

    **tags**            --  returns the tags present in this tagset

    **tag_set_id**      --  returns the tagset id

Tag:

    __init__()                  --  Initialise object of the Tag class

    _response_not_success()     --  parses through the exception response, and raises SDKException

    _get_tag_id()               --  Returns the tag id of the given tag name

    _get_tag_properties()       --  Returns the properties of Tag

    refresh()                   --  refresh the tag details

    modify()                    --  modifies the name of the tag

Tag Attributes
-----------------

    **guid**            --  returns tag GUID of this tag

    **full_name**       --  returns the full name of this tag

    **tag_id**          --  returns the id of the tag


ActivateEntities:

    __init__(commcell_object)           --  initialise object of the ActivateEntities class

     _response_not_success()            --  parses through the exception response, and raises SDKException

    refresh()                           --  refresh the regex entities associated with the commcell

    get()                               --  Returns an instance of ActivateEntity class for the given regex entity name

    get_entity_ids()                    --  Returns an list of entity ids for the given regex entity name list

    get_entity_keys()                   --  Returns an list of entity keys for the given regex entity name list

    get_properties()                    --  Returns the properties for the given regex entity name

    _get_all_activate_entities()        --  Returns dict consisting all regex entities associated with commcell

    _get_regex_entity_from_collections()--  gets all the regex entity details from collection response

    has_entity()                        --  Checks whether given regex entity exists in commcell or not

    add()                               --  adds the regex entity in the commcell

    delete()                            --  deletes the regex entity in the commcell for given entity name

    _process_entity_containers()        -- returns the container details for the entity

ActivateEntity:

    __init__(
        commcell_object,
        entity_name,
        entity_id=None)             --  initialize an object of ActivateEntity Class with the given regex entity
                                                name and id

     _response_not_success()            --  parses through the exception response, and raises SDKException

    refresh()                           --  refresh the properties of the regex entity

    _get_entity_id()                    --  Gets entity id for the given regex entity name

    _get_entity_properties()            --  Gets all the details of associated regex entity

    modify()                            --  Modifies the entity properties for the associated regex entity


ActivateEntity Attributes
-----------------

    **entity_id**         --  returns the id of the regex entity

    **entity_key**        --  returns the key of the regex entity

    **category_name**     --  returns the category name of the regex entity

    **is_enabled**        --  returns the enabled flag of the regex entity

    **display_name**      --  returns the display name of the regex entity

    **entity_type**       --  returns the type of entity (1- NER 2-RER 3-Derived 4-Classifier)

    **entity_xml**        --  returns the entity xml associated with this entity

    **container_details** --  returns the container details for this entity

Classifiers:

    __init__(commcell_object)           --  initialise object of the Classifiers class

     _response_not_success()            --  parses through the exception response, and raises SDKException

    refresh()                           --  refresh the classifiers associated with the commcell

    get()                               --  Returns an instance of Classifier class for the given classifier name

    get_entity_ids()                    --  Returns an list of entity ids for the given classifier name list

    get_entity_keys()                   --  Returns an list of entity keys for the given classifier name list

    get_properties()                    --  Returns the properties for the given classifier name

    _get_all_classifier_entities()      --  Returns dict consisting all classifier associated with commcell

    _get_classifier_entity_from_collections()-- gets all the classifier details from collection response

    has_classifier()                    --  Checks whether given classifier exists in commcell or not

    add()                               --  adds the classifier in the commcell

    delete()                            --  deletes the classifier in the commcell



Classifier:

    __init__(
        commcell_object,
        classifier_name,
        entity_id=None)             --  initialize an object of Classifier Class with the given classifier
                                                name and id

     _response_not_success()            --  parses through the exception response, and raises SDKException

    refresh()                           --  refresh the properties of the classifier

    _get_entity_id()                    --  Gets entity id for the given classifier name

    _get_entity_properties()            --  Gets all the properties of associated classifier

    _get_upload_request_id()            --  returns the request id for uploading model data

    _get_upload_api()                   --  returns the upload API url

    _validate_upload_response()         --  Validates the upload api response for given chunk size

    _get_upload_flag_bit()              --  Returns the 7-byte flag used in upload API call

    modify()                            --  Modifies the properties for the associated classifier

    upload_data()                       --  Uploads the training model data for this classifier

    start_training()                    --  Starts training for this classifier

    monitor_training()                  --  Monitors training status for this classifier

    cancel_training()                   --  Cancels training for this classifier


Classifier Attributes
-----------------

    **entity_id**               --  returns the id of the classifier

    **entity_key**              --  returns the key of the classifier

    **category_name**           --  returns the category name of the classifier

    **is_enabled**              --  returns the enabled flag of the classifier

    **display_name**            --  returns the display name of the classifier

    **entity_type**             --  returns the type of entity (4-Classifier)

    **entity_xml**              --  returns the entity xml associated with this classifier

    **trained_ca_cloud_id**     -- returns the cloud id of the trained content analyzer

    **training_status**         --  returns the training status for this classifier

    **sycn_ca_client_id**       --  returns the list of synced CA client id for this classifier

    **last_training_time**      --  returns the last training time for this classifier

    **training_accuracy**       --  returns the training accuracy for this classifier

    **sample_details**          --  returns dict containing model sample count details used for this classifier training

"""
import copy
import os
import time
from enum import Enum

from ..exception import SDKException
from .constants import ActivateEntityConstants, ClassifierConstants, TrainingStatus
from .constants import TagConstants


class EntityManagerTypes(Enum):
    """Class to represent different entity types in entity manager"""
    ENTITIES = "Entities"
    CLASSIFIERS = "Classifiers"
    TAGS = "Tags"


class ActivateEntities(object):
    """Class for representing all the regex entities in the commcell."""

    def __init__(self, commcell_object):
        """Initializes an instance of the ActivateEntities class.

            Args:
                commcell_object     (object)    --  instance of the commcell class

            Returns:
                object  -   instance of the ActivateEntities class

        """
        self._commcell_object = commcell_object
        self._update_response_ = commcell_object._update_response_
        self._cvpysdk_object = commcell_object._cvpysdk_object
        self._services = commcell_object._services
        self._regex_entities = None
        self._entities_containers = None
        self._api_get_all_regex_entities = self._services['ACTIVATE_ENTITIES']
        self._api_get_containers = self._services['ACTIVATE_ENTITY_CONTAINER']
        self._api_create_regex_entity = self._api_get_all_regex_entities
        self._api_delete_regex_entity = self._services['ACTIVATE_ENTITY']
        self.refresh()

    def add(self, entity_name, entity_regex, entity_keywords, entity_flag, is_derived=False, parent_entity=None):
        """Adds the specified regex entity name in the commcell

                    Args:
                        entity_name (str)      --  name of the regex entity

                        entity_regex (str)     --  Regex for the entity

                        entity_keywords (str)  --  Keywords for the entity

                        entity_flag (int)      --  Sensitivity flag value for entity
                                                        5-Highly sensitive
                                                        3-Moderate sensitive
                                                        1-Low sensitive

                        is_derived (bool)      --  represents whether it is derived entity or not

                        parent_entity(int)     -- entity id of the parent entity in case of derived entity

                    Returns:
                        None

                    Raises:
                        SDKException:

                                if response is empty

                                if response is not success

                                if unable to add regex entity in commcell

                                if entity_flag is not in proper allowed values [1,3,5]

                                if input data type is not valid
                """
        if not isinstance(entity_name, str) or not isinstance(entity_regex, str) \
                or not isinstance(entity_keywords, str):
            raise SDKException('ActivateEntity', '101')
        if entity_flag not in [1, 3, 5]:
            raise SDKException('ActivateEntity', '102', 'Unsupported entity flag value')
        request_json = ActivateEntityConstants.REQUEST_JSON
        request_json['regularExpression'] = "{\"entity_key\":\"" + \
            entity_name + "\",\"entity_regex\":\"" + entity_regex + "\"}"
        request_json['flags'] = entity_flag
        request_json['entityName'] = entity_name
        request_json['entityXML']['keywords'] = entity_keywords
        if is_derived:
            request_json['regularExpression'] = "{\"entity_key\":\"" + entity_name + "\"}"
            request_json['entityType'] = 3
            if isinstance(parent_entity, int):
                request_json['parentEntityId'] = parent_entity
            elif isinstance(parent_entity, str):
                request_json['parentEntityId'] = self._regex_entities[parent_entity]['entityId']
            else:
                raise SDKException('ActivateEntity', '102', 'Unsupported parent entity id type provided')

        flag, response = self._cvpysdk_object.make_request(
            'POST', self._api_create_regex_entity, request_json
        )

        if flag:
            if response.json() and 'entityDetails' in response.json() and 'err' not in response.json():
                self.refresh()
                return
            raise SDKException('ActivateEntity', '104')
        self._response_not_success(response)

    def delete(self, entity_name):
        """deletes the specified regex entity name in the commcell

                    Args:
                        entity_name (str)      --  name of the regex entity

                    Returns:
                        None

                    Raises:
                        SDKException:

                                if response is empty

                                if response is not success

                                if unable to delete regex entity in commcell

                                if unable to find entity name in the commcell

                                if data type of entity_name is invalid


                """
        if not isinstance(entity_name, str):
            raise SDKException('ActivateEntity', '101')
        if entity_name not in self._regex_entities:
            raise SDKException('ActivateEntity', '102', 'Unable to find given regex entity name in the commcell')

        flag, response = self._cvpysdk_object.make_request(
            'DELETE', self._api_delete_regex_entity % self._regex_entities[entity_name]['entityId']
        )

        if flag:
            if response.json() and 'errorCode' in response.json() and response.json()['errorCode'] == 0:
                self.refresh()
                return
            raise SDKException('ActivateEntity', '105')
        self._response_not_success(response)

    def _response_not_success(self, response):
        """Helper function to raise an exception when reponse status is not 200 (OK).

            Args:
                response    (object)    --  response class object,

                received upon running an API request, using the `requests` python package

        """
        raise SDKException('Response', '101', self._update_response_(response.text))

    def get_properties(self, entity_name):
        """Returns a properties of the specified regex entity name.

            Args:
                entity_name (str)  --  name of the regex entity

            Returns:
                dict -  properties for the given regex entity name


        """
        return self._regex_entities[entity_name]

    def _get_all_activate_entities(self):
        """Gets the list of all regex entities associated with this commcell.

            Returns:
                dict    -   dictionary consisting of dictionaries, where each dictionary stores the
                                details of a single regex entity

                    {
                        "entityDetails": [
                            {
                                "displayName": "US Social Security number",
                                "flags": 5,
                                "description": "",
                                "categoryName": "US",
                                "enabled": true,
                                "entityName": "SSN",
                                "attribute": 3,
                                "entityType": 2,
                                "entityKey": "ssn",
                                "entityId": 1111,
                                "entityXML": {
                                    "keywords": "Social Security,Social Security#,Soc Sec,SSN,SSNS,SSN#,SS#,SSID",
                                    "entityKey": "ssn",
                                    "isSystemDefinedEntity": true,
                                    "inheritBaseWords": false
                                }
                            },
                            {
                                "displayName": "Person Name",
                                "flags": 1,
                                "description": "Name of a person.",
                                "categoryName": "Generic",
                                "enabled": true,
                                "entityName": "Person",
                                "attribute": 3,
                                "entityType": 1,
                                "entityKey": "person",
                                "entityId": 1112,
                                "entityXML": {
                                    "keywords": "",
                                    "entityKey": "person",
                                    "inheritBaseWords": false
                                }
                            }
                            ]
                            }

            Raises:
                SDKException:
                    if response is empty

                    if response is not success

        """
        flag, response = self._cvpysdk_object.make_request(
            'GET', self._api_get_containers
        )
        if flag:
            if response.json() and 'containerTypesList' in response.json():
                self._entities_containers = response.json()['containerTypesList']
            else:
                raise SDKException('ActivateEntity', '107')
        else:
            self._response_not_success(response)

        flag, response = self._cvpysdk_object.make_request(
            'GET', self._api_get_all_regex_entities
        )

        if flag:
            if response.json() and 'entityDetails' in response.json():
                return self._get_regex_entity_from_collections(response.json())
            raise SDKException('ActivateEntity', '103')
        self._response_not_success(response)

    def _process_entity_containers(self, entity_name, container_name):
        """Returns container details for given entity name & container name

            Args:

                entity_name         (str)       --  Entity name

                container_name      (str)       --  Container name

            Returns:

                dict    --  Container details of entity
        """
        output = {}
        for dept in self._entities_containers:
            items_list = dept['tagSetsAndItems']
            for country in items_list:
                tags_list = country.get('tags',{})
                for tag in tags_list:
                    if entity_name.lower() == tag['entityDetail']['entityName'].lower() and \
                            container_name.lower() == country['container']['containerName'].lower():
                        output['tags'] = [tag]
                        output['container'] = country['container']
                        return output
        return output

    def _get_regex_entity_from_collections(self, collections):
        """Extracts all the regex entities, and their details from the list of collections given,
            and returns the dictionary of all regex entities

            Args:
                collections     (list)  --  list of all collections

            Returns:
                dict    -   dictionary consisting of dictionaries, where each dictionary stores the
                                details of a single regex entity

        """
        _regex_entity = {}
        for regex_entity in collections['entityDetails']:
            regex_entity_dict = {}
            regex_entity_dict['displayName'] = regex_entity.get('displayName', "")
            regex_entity_dict['entityKey'] = regex_entity.get('entityKey', "")
            regex_entity_dict['categoryName'] = regex_entity.get('categoryName', "")
            if regex_entity_dict['categoryName'] is not None:
                regex_entity_dict['containerDetails'] = self._process_entity_containers(
                    entity_name=regex_entity_dict['displayName'], container_name=regex_entity_dict['categoryName'])
            regex_entity_dict['entityXML'] = regex_entity.get('entityXML', "")
            regex_entity_dict['entityId'] = regex_entity.get('entityId', 0)
            regex_entity_dict['flags'] = regex_entity.get('flags', 0)
            regex_entity_dict['entityType'] = regex_entity.get('entityType', 0)
            regex_entity_dict['enabled'] = regex_entity.get('enabled', False)
            _regex_entity[regex_entity['entityName']] = regex_entity_dict
        return _regex_entity

    def refresh(self):
        """Refresh the activate regex entities associated with the commcell."""
        self._regex_entities = self._get_all_activate_entities()

    def get(self, entity_name):
        """Returns a ActivateEntity object for the given regex entity name.

            Args:
                entity_name (str)  --  name of the regex entity

            Returns:

                obj                 -- Object of ActivateEntity class

            Raises:
                SDKException:
                    if response is empty

                    if response is not success

                    if entity_name is not of type string


        """
        if not isinstance(entity_name, str):
            raise SDKException('ActivateEntity', '101')

        if self.has_entity(entity_name):
            entity_id = self._regex_entities[entity_name]['entityId']
            return ActivateEntity(self._commcell_object, entity_name, entity_id)
        raise SDKException('ActivateEntity', '102', "Unable to get ActivateEntity class object")

    def get_entity_ids(self, entity_name):
        """Returns a list of entity id for the given regex entity name list.

            Args:
                entity_name (list)  --  names of the regex entity

            Returns:

                list                -- entity id's for the given entity names


        """
        if not isinstance(entity_name, list):
            raise SDKException('ActivateEntity', '101')
        entity_ids = []
        for regex_entity in entity_name:
            if regex_entity in self._regex_entities:
                entity_ids.append(self._regex_entities[regex_entity]['entityId'])
            else:
                raise SDKException(
                    'ActivateEntity', '102', f"Unable to find entity id for given entity name :{regex_entity}")
        return entity_ids

    def get_entity_keys(self, entity_name):
        """Returns a list of entity keys for the given regex entity name list.

            Args:
                entity_name (list)  --  names of the regex entity

            Returns:

                list                -- entity keys for the given entity names


        """
        if not isinstance(entity_name, list):
            raise SDKException('ActivateEntity', '101')
        entity_keys = []
        for regex_entity in entity_name:
            if regex_entity in self._regex_entities:
                entity_keys.append(self._regex_entities[regex_entity]['entityKey'])
            else:
                raise SDKException(
                    'ActivateEntity', '102', f"Unable to find entity keys for given entity name :{regex_entity}")
        return entity_keys

    def has_entity(self, entity_name):
        """Checks if a regex entity exists in the commcell with the input name.

            Args:
                entity_name (str)  --  name of the regex entity

            Returns:
                bool - boolean output whether the regex entity exists in the commcell or not

            Raises:
                SDKException:
                    if type of the regex entity name argument is not string

        """
        if not isinstance(entity_name, str):
            raise SDKException('ActivateEntity', '101')

        return self._regex_entities and entity_name.lower() in map(str.lower, self._regex_entities)


class ActivateEntity(object):
    """Class for performing operations on a single regex entity"""

    def __init__(self, commcell_object, entity_name, entity_id=None):
        """Initialize an object of the ActivateEntity class.

            Args:
                commcell_object     (object)    --  instance of the commcell class

                entity_name     (str)           --  name of the regex entity

                entity_id       (str)           --  id of the regex entity
                    default: None

            Returns:
                object  -   instance of the ActivateEntity class
        """
        self._commcell_object = commcell_object
        self._update_response_ = commcell_object._update_response_
        self._services = commcell_object._services
        self._cvpysdk_obj = self._commcell_object._cvpysdk_object
        self._entity_name = entity_name
        self._entity_id = None
        self._display_name = None
        self._entity_type = None
        self._is_enabled = None
        self._entity_key = None
        self._entity_xml = None
        self._category_name = None
        self._container_details = None
        if entity_id is None:
            self._entity_id = self._get_entity_id(entity_name)
        else:
            self._entity_id = entity_id
        self.refresh()
        self._api_modify_regex_entity = self._services['ACTIVATE_ENTITY']

    def _response_not_success(self, response):
        """Helper function to raise an exception when reponse status is not 200 (OK).

            Args:
                response    (object)    --  response class object,

                received upon running an API request, using the `requests` python package

        """
        raise SDKException('Response', '101', self._update_response_(response.text))

    def modify(self, entity_regex, entity_keywords, entity_flag, is_derived=False, parent_entity=None):
        """Modifies the specified regex entity details

                    Args:

                        entity_regex (str)     --  Regex for the entity

                        entity_keywords (str)  --  Keywords for the entity

                        entity_flag (int)      --  Sensitivity flag value for entity
                                                        5-Highly sensitive
                                                        3-Moderate sensitive
                                                        1-Low sensitive

                        is_derived (bool)      --  represents whether it is derived entity or not

                        parent_entity(int)     -- entity id of the parent entity in case of derived entity

                    Returns:
                        None

                    Raises:
                        SDKException:

                                if response is empty

                                if response is not success

                                if unable to modify regex entity in commcell

                                if input entity_keywords & entity_regex is not string

                                if entity_flag value is not in allowed values[1,3,5]


                """
        if not isinstance(entity_regex, str) \
                or not isinstance(entity_keywords, str):
            raise SDKException('ActivateEntity', '101')
        if entity_flag not in [1, 3, 5]:
            raise SDKException('ActivateEntity', '102', 'Unsupported entity flag value')
        request_json = ActivateEntityConstants.REQUEST_JSON
        request_json['regularExpression'] = "{\"entity_key\":\"" + \
            self._entity_name + "\",\"entity_regex\":\"" + entity_regex + "\"}"
        request_json['flags'] = entity_flag
        request_json['entityName'] = self._entity_name
        request_json['entityXML']['keywords'] = entity_keywords
        if is_derived:
            request_json['regularExpression'] = "{\"entity_key\":\"" + self._entity_name + "\"}"
            request_json['entityType'] = 3
            if isinstance(parent_entity, int):
                request_json['parentEntityId'] = parent_entity
            elif isinstance(parent_entity, str):
                request_json['parentEntityId'] = ActivateEntities.get(
                    self._commcell_object, entity_name=parent_entity).entity_id
            else:
                raise SDKException('ActivateEntity', '102', 'Unsupported parent entity id type provided')

        flag, response = self._cvpysdk_obj.make_request(
            'PUT', (self._api_modify_regex_entity % self.entity_id), request_json
        )

        if flag:
            if response.json() and 'entityDetails' in response.json() and 'err' not in response.json():
                return
            raise SDKException('ActivateEntity', '106')
        self._response_not_success(response)

    def _get_entity_id(self, entity_name):
        """ Get regex entity id for given entity name
                Args:

                    entity_name (str)  -- Name of the regex entity

                Returns:

                    int                -- id of the regex entity

        """

        return self._commcell_object.activate.entity_manager().get(entity_name).entity_id

    def _get_entity_properties(self):
        """ Get regex entity properties for given entity name
                Args:

                    None

                Returns:

                    None

        """

        regex_entity_dict = self._commcell_object.activate.entity_manager().get_properties(self._entity_name)
        self._display_name = regex_entity_dict['displayName']
        self._category_name = regex_entity_dict['categoryName']
        self._entity_id = regex_entity_dict['entityId']
        self._is_enabled = regex_entity_dict['enabled']
        self._entity_key = regex_entity_dict['entityKey']
        self._entity_type = regex_entity_dict['entityType']
        self._entity_xml = regex_entity_dict['entityXML']
        self._container_details = regex_entity_dict['containerDetails']
        return regex_entity_dict

    @property
    def container_details(self):
        """Returns the container details for this entity"""
        return self._container_details

    @property
    def entity_id(self):
        """Returns the value of the regex entity id attribute."""
        return self._entity_id

    @property
    def entity_type(self):
        """Returns the entity type attribute."""
        return self._entity_type

    @property
    def category_name(self):
        """Returns the entity category name attribute."""
        return self._category_name

    @property
    def display_name(self):
        """Returns the entity display name attribute."""
        return self._display_name

    @property
    def is_enabled(self):
        """Returns the entity isenabled attribute."""
        return self._is_enabled

    @property
    def entity_key(self):
        """Returns the entity key attribute."""
        return self._entity_key

    @property
    def entity_xml(self):
        """Returns the entity xml attribute."""
        return self._entity_xml

    def refresh(self):
        """Refresh the regex entity details for associated object"""
        self._get_entity_properties()


class Tags(object):
    """Class for representing all the Tagsets in the commcell."""

    def __init__(self, commcell_object):
        """Initializes an instance of the Tags class.

            Args:
                commcell_object     (object)    --  instance of the commcell class

            Returns:
                object  -   instance of the Tags class

        """
        self._commcell_object = commcell_object
        self._update_response_ = commcell_object._update_response_
        self._cvpysdk_object = commcell_object._cvpysdk_object
        self._services = commcell_object._services
        self._tag_set_entities = None
        self._api_get_all_tag_sets = self._services['GET_TAGS']
        self._api_add_tag_set = self._services['ADD_CONTAINER']
        self._api_delete_tag_set = self._services['DELETE_CONTAINER']
        self.refresh()

    def _response_not_success(self, response):
        """Helper function to raise an exception when reponse status is not 200 (OK).

            Args:
                response    (object)    --  response class object,

                received upon running an API request, using the `requests` python package

        """
        raise SDKException('Response', '101', self._update_response_(response.text))

    @staticmethod
    def _get_tag_sets_from_collections(collections):
        """Extracts all the tagsets, and their details from the list of collections given,
            and returns the dictionary of all tagsets

            Args:
                collections     (list)  --  list of all tagsets

            Returns:
                dict    -   dictionary consisting of dictionaries, where each dictionary stores the
                                details of a single tagset

        """
        _tag_set_entity = {}
        for tagset in collections['listOftagSetList']:
            tagset = tagset['tagSetsAndItems']
            container = tagset[0]['container']
            owner_info = tagset[0]['container']['ownerInfo']
            tagset_dict = {}
            tagset_dict['containerName'] = container.get('containerName', "")
            tagset_dict['containerFullName'] = container.get('containerFullName', "")
            tagset_dict['containerId'] = container.get('containerId', "")
            tagset_dict['containerGuid'] = container.get('containerGuid', "")
            tagset_dict['comment'] = container.get('comment', "")
            tagset_dict['owneruserName'] = owner_info.get('userName', "")
            tagset_dict['owneruserGuid'] = owner_info.get('userGuid', "")
            tagset_dict['owneraliasName'] = owner_info.get('aliasName', "")
            container_tags = []
            tag_ids = []
            tag_dict = {}
            # process tags only if it is present
            if 'tags' in tagset[0]:
                tags = tagset[0]['tags']
                for tag in tags:
                    tag_dict[tag['name'].lower()] = tag
                    container_tags.append(tag['name'].lower())
                    tag_ids.append(tag['tagId'])
            tagset_dict['tags'] = container_tags
            tagset_dict['tagsIds'] = tag_ids
            tagset_dict['tagsDetails'] = tag_dict
            _tag_set_entity[tagset_dict['containerName'].lower()] = tagset_dict
        return _tag_set_entity

    def _get_all_tag_sets(self):
        """Gets the list of all tagsets associated with this commcell.

            Returns:
                dict    -   dictionary consisting of dictionaries, where each dictionary stores the
                                details of a single tagset entity

            Raises:
                SDKException:
                    if response is empty

                    if response is not success

        """
        flag, response = self._cvpysdk_object.make_request(
            'GET', self._api_get_all_tag_sets
        )

        if flag:
            if response.json() and 'listOftagSetList' in response.json():
                return self._get_tag_sets_from_collections(response.json())
            raise SDKException('Tags', '103')
        self._response_not_success(response)

    def get(self, tag_set_name):
        """Returns a TagSet object for the given Tagset name.

            Args:
                tag_set_name (str)  --  name of the TagSet

            Returns:

                obj                 -- Object of TagSet class

            Raises:
                SDKException:
                    if response is empty

                    if response is not success

                    if tag_set_name is not of type string


        """
        if not isinstance(tag_set_name, str):
            raise SDKException('Tags', '101')

        if self.has_tag_set(tag_set_name):
            tag_set_id = self._tag_set_entities[tag_set_name.lower()]['containerId']
            return TagSet(self._commcell_object, tag_set_name=tag_set_name, tag_set_id=tag_set_id)
        raise SDKException('Tags', '102', "Unable to get TagSet class object")

    def has_tag_set(self, tag_set_name):
        """Checks if a tagset exists in the commcell with the input name or not

            Args:
                tag_set_name (str)  --  name of the TagSet

            Returns:
                bool - boolean output whether the TagSet exists in the commcell or not

            Raises:
                SDKException:
                    if type of the TagSet name argument is not string

        """
        if not isinstance(tag_set_name, str):
            raise SDKException('Tags', '101')
        return self._tag_set_entities and tag_set_name.lower() in map(str.lower, self._tag_set_entities)

    def get_properties(self, tag_set_name):
        """Returns a properties of the specified TagSet name.

            Args:
                tag_set_name (str)  --  name of the TagSet

            Returns:
                dict -  properties for the given TagSet name


                Example : {
                              "containerName": "cvpysdk1",
                              "containerFullName": "cvpysdk1",
                              "containerId": 65931,
                              "containerGuid": "6B870271-543A-4B76-955D-CDEB3807D68E",
                              "comment": "Created from CvPySDK",
                              "owneruserName": "xxx",
                              "owneruserGuid": "C31C1194-AA5C-47C3-B5B0-9087EF429B6B",
                              "owneraliasName": "xx",
                              "tags": [
                                "p10"
                              ],
                              "tagsIds": [
                                15865
                              ],
                              "tagsDetails": {
                                "p10": {
                                  "tagOwnerType": 1,
                                  "tagId": 15865,
                                  "name": "p10",
                                  "flags": 0,
                                  "fullName": "cvpysdk1\\p10",
                                  "description": "",
                                  "id": "C9E229D0-B895-4653-9DA7-C9C6BD999121",
                                  "attribute": {}
                                }
                              }
                            }


        """
        return self._tag_set_entities[tag_set_name.lower()]

    def delete(self, tag_set_name):
        """Deletes the specified tagset from the commcell

                Args:

                    tag_set_name    (str)       --  Name of the Tagset

                Returns:

                    None

                Raises:

                    SDKException:

                                if response is empty

                                if response is not success

                                if unable to delete TagSet entity in commcell

                                if input data type is not valid

                                if unable to find TagSet entity in commcell
        """
        if not isinstance(tag_set_name, str):
            raise SDKException('Tags', '101')
        if not self.has_tag_set(tag_set_name=tag_set_name):
            raise SDKException('Tags', '102', "Tagset not found")
        request_json = copy.deepcopy(TagConstants.TAG_SET_DELETE_REQUEST_JSON)
        request_json['containers'][0]['containerId'] = self._tag_set_entities[tag_set_name.lower()]['containerId']
        flag, response = self._cvpysdk_object.make_request(
            'POST', self._api_delete_tag_set, request_json
        )
        if flag:
            if response.json() and 'errorCode' in response.json():
                if int(response.json()['errorCode']) != 0:
                    raise SDKException('Tags', '102', response.json()['errorMessage'])
            elif 'errList' in response.json():
                raise SDKException('Tags', '102', response.json()['errList'][0]['errLogMessage'])
            self.refresh()
            return
        self._response_not_success(response)

    def add(self, tag_set_name, comment="Created from CvPySDK"):
        """Adds the specified TagSet name in the commcell

                    Args:
                        tag_set_name (str)     --  name of the TagSet

                        comment (str)         --  Comment for this TagSet

                    Returns:

                        object      --  Object of TagSet class

                    Raises:
                        SDKException:

                                if response is empty

                                if response is not success

                                if unable to add TagSet entity in commcell

                                if input data type is not valid
                """
        if not isinstance(tag_set_name, str) or not isinstance(comment, str):
            raise SDKException('Tags', '101')
        request_json = copy.deepcopy(TagConstants.TAG_SET_ADD_REQUEST_JSON)
        request_json['container']['containerName'] = tag_set_name
        request_json['container']['comment'] = comment
        flag, response = self._cvpysdk_object.make_request(
            'POST', self._api_add_tag_set, request_json
        )
        if flag:
            if response.json():
                if 'errList' in response.json():
                    raise SDKException('Tags', '102', response.json()['errList'][0]['errLogMessage'])
                elif 'container' in response.json():
                    self.refresh()
                    return TagSet(commcell_object=self._commcell_object, tag_set_name=tag_set_name)
            raise SDKException('Tags', '104')
        self._response_not_success(response)

    def refresh(self):
        """Refresh the TagSet entities associated with the commcell."""
        self._tag_set_entities = self._get_all_tag_sets()


class TagSet(object):
    """Class for performing operations on a TagSet"""

    def __init__(self, commcell_object, tag_set_name, tag_set_id=None):
        """Initialize an object of the TagSet class.

            Args:
                commcell_object     (object)    --  instance of the commcell class

                tag_set_name     (str)          --  name of the TagSet

                tag_set_id       (str)          --  Container id of the TagSet
                                                        default: None

            Returns:
                object  -   instance of the Tagset class
        """
        self._commcell_object = commcell_object
        self._update_response_ = commcell_object._update_response_
        self._services = commcell_object._services
        self._cvpysdk_obj = self._commcell_object._cvpysdk_object
        self._tag_set_name = tag_set_name
        self._tag_set_id = None
        self._tag_set_props = None
        if tag_set_id is None:
            self._tag_set_id = self._get_tag_set_id(tag_set_name)
        else:
            self._tag_set_id = tag_set_id
        self._container_guid = None
        self._owner = None
        self._full_name = None
        self._comment = None
        self._tags = None
        self._tag_ids = None
        self._owner_alias_name = None
        self._api_modify_tag_set = self._services['ADD_CONTAINER']
        self._api_add_tag = self._services['GET_TAGS']
        self._api_security = self._services['SECURITY_ASSOCIATION']
        self.refresh()

    def _response_not_success(self, response):
        """Helper function to raise an exception when reponse status is not 200 (OK).

            Args:
                response    (object)    --  response class object,

                received upon running an API request, using the `requests` python package

        """
        raise SDKException('Response', '101', self._update_response_(response.text))

    def has_tag(self, tag_name):
        """Returns whether tag exists with given name or not in tagset

                    Args:
                        tag_name (str)      --  name of the Tag

                    Returns:

                        bool    --  True if it exists or else false

                    Raises:
                        SDKException:

                            if tag_name is not of type string


        """
        if not isinstance(tag_name, str):
            raise SDKException('Tags', '101')
        if tag_name.lower() in self.tags:
            return True
        return False

    def get(self, tag_name):
        """Returns a Tag object for the given Tag name.

            Args:
                tag_name (str)      --  name of the Tag

            Returns:

                obj                 -- Object of Tag class

            Raises:
                SDKException:

                    if unable to create Tag object

                    if tag_name is not of type string


        """
        if not isinstance(tag_name, str):
            raise SDKException('Tags', '101')
        if self.has_tag(tag_name):
            return Tag(self._commcell_object, tag_set_name=self._tag_set_name, tag_name=tag_name)
        raise SDKException('Tags', '102', "Unable to get Tag class object")

    def get_tag_id(self, tag_name):
        """Returns the tag id for the given tag name

                Args:

                    tag_name        (str)       --  Name of the tag

                Returns:

                    int     --  Tag id

                Raises:

                    SDKExeption:

                        if input tag name is not found in this tagset

        """
        if tag_name.lower() not in self.tags:
            raise SDKException('Tags', '106')
        index = self.tags.index(tag_name.lower())
        return self._tag_ids[index]

    def add_tag(self, tag_name):
        """Adds the specified tag name in the tagset container in commcell

                           Args:
                               tag_name (str)     --  name of the Tag

                           Returns:

                               object      --  Object of Tag class

                           Raises:
                               SDKException:

                                       if response is empty

                                       if response is not success

                                       if unable to add Tag inside Tagset in commcell

                                       if input data type is not valid
                       """
        if not isinstance(tag_name, str):
            raise SDKException('Tags', '101')
        request_json = copy.deepcopy(TagConstants.TAG_ADD_REQUEST_JSON)
        request_json['container']['containerId'] = self._tag_set_id
        request_json['tags'][0]['name'] = tag_name
        flag, response = self._cvpysdk_obj.make_request(
            'POST', self._api_add_tag, request_json
        )
        if flag:
            if response.json():
                if 'errList' in response.json():
                    raise SDKException('Tags', '102', response.json()['errList'][0]['errLogMessage'])
                elif 'tag' in response.json():
                    self.refresh()
                    return Tag(commcell_object=self._commcell_object, tag_set_name=self._tag_set_name,
                               tag_name=tag_name)
            raise SDKException('Tags', '104')
        self._response_not_success(response)

    def share(self, user_or_group_name, allow_edit_permission=False, is_user=True, ops_type=1):
        """Shares tagset with given user or group in commcell

                Args:

                    user_or_group_name      (str)       --  Name of user or group

                    is_user                 (bool)      --  Denotes whether this is user or group name
                                                                default : True(User)

                    allow_edit_permission   (bool)      --  whether to give edit permission or not to user or group

                    ops_type                (int)       --  Operation type

                                                            Default : 1 (Add)

                                                            Supported : 1 (Add)
                                                                        2 (Modify)
                                                                        3 (Delete)

                Returns:

                    None

                Raises:

                    SDKException:

                            if unable to update security associations

                            if response is empty or not success
        """
        if not isinstance(user_or_group_name, str):
            raise SDKException('Tags', '101')
        request_json = copy.deepcopy(TagConstants.TAG_SET_SHARE_REQUEST_JSON)
        external_user = False
        if '\\' in user_or_group_name:
            external_user = True

        if is_user:
            user_obj = self._commcell_object.users.get(user_or_group_name)
            user_id = user_obj.user_id
            user_or_group_name = f"\\{user_or_group_name}"
            request_json['securityAssociations']['associations'][0]['userOrGroup'][0]['userId'] = int(user_id)
            request_json['securityAssociations']['associations'][0]['userOrGroup'][0]['_type_'] = "13"
            request_json['securityAssociations']['associations'][0]['userOrGroup'][0]['userName'] = user_or_group_name
        elif external_user:
            request_json['securityAssociations']['associations'][0]['userOrGroup'][0]['groupId'] = 0
            request_json['securityAssociations']['associations'][0]['userOrGroup'][0]['_type_'] = "62"
            request_json['securityAssociations']['associations'][0]['userOrGroup'][0]['externalGroupName'] = user_or_group_name
        else:
            grp_obj = self._commcell_object.user_groups.get(user_or_group_name)
            grp_id = grp_obj.user_group_id
            request_json['securityAssociations']['associations'][0]['userOrGroup'][0]['userGroupId'] = int(grp_id)
            request_json['securityAssociations']['associations'][0]['userOrGroup'][0]['_type_'] = "15"
            request_json['securityAssociations']['associations'][0]['userOrGroup'][0]['userGroupName'] = user_or_group_name

        request_json['entityAssociated']['entity'][0]['tagId'] = self._tag_set_id
        request_json['securityAssociations']['associationsOperationType'] = ops_type

        if allow_edit_permission:
            request_json['securityAssociations']['associations'][0]['properties']['permissions'].append(
                TagConstants.ADD_PERMISSION)
        flag, response = self._cvpysdk_obj.make_request(
            'POST', self._api_security, request_json
        )
        if flag:
            if response.json() and 'response' in response.json():
                response_json = response.json()['response'][0]
                error_code = response_json['errorCode']
                if error_code != 0:
                    error_message = response_json['errorString']
                    raise SDKException(
                        'Tags',
                        '102', error_message)
            else:
                raise SDKException('Tags', '107')
        else:
            response_string = self._commcell_object._update_response_(response.text)
            raise SDKException('Response', '101', response_string)

    def modify(self, new_name=None, comment="Modified from CvPySDK"):
        """Modifies the specified tagset in the commcell

                Args:

                    new_name        (str)       --  New name for Tagset

                    comment         (str)       --  New comment which needs to be added for Tagset

                Returns:

                    None

                Raises:

                    SDKException:

                                if response is empty

                                if response is not success

                                if unable to modify TagSet entity in commcell

                                if input is not a valid data type of string

        """
        if not isinstance(new_name, str) or not isinstance(comment, str):
            raise SDKException('Tags', '101')
        request_json = copy.deepcopy(TagConstants.TAG_SET_MODIFY_REQUEST_JSON)
        request_json['container']['containerName'] = new_name
        request_json['container']['comment'] = comment
        request_json['container']['containerId'] = self._tag_set_id
        flag, response = self._cvpysdk_obj.make_request(
            'POST', self._api_modify_tag_set, request_json
        )
        if flag:
            if response.json():
                if 'errList' in response.json():
                    raise SDKException('Tags', '102', response.json()['errList'][0]['errLogMessage'])
                elif 'container' in response.json():
                    self._tag_set_name = new_name
                    self.refresh()
                    return
            raise SDKException('Tags', '105')
        self._response_not_success(response)

    def refresh(self):
        """Refresh the TagSet details for associated object"""
        self._tag_set_props = self._get_tag_set_properties()

    def _get_tag_set_properties(self):
        """ Get TagSet properties for this associated object
                Args:

                    None

                Returns:

                    dict    --  Containing tagset properties

        """
        tags = self._commcell_object.activate.entity_manager(
            EntityManagerTypes.TAGS)
        # call refresh before fetching properties
        tags.refresh()
        tag_set_dict = tags.get_properties(self._tag_set_name)
        self._full_name = tag_set_dict['containerFullName']
        self._owner = tag_set_dict['owneruserName']
        self._comment = tag_set_dict['comment']
        self._container_guid = tag_set_dict['containerGuid']
        self._tags = tag_set_dict['tags']
        self._tag_ids = tag_set_dict['tagsIds']
        self._owner_alias_name = tag_set_dict['owneraliasName']
        return tag_set_dict

    def _get_tag_set_id(self, tag_set_name):
        """ Get TagSet container id for given tag set name
                Args:

                    tag_set_name (str)  -- Name of the TagSet

                Returns:

                    int                --  TagSet container Id

        """

        return self._commcell_object.activate.entity_manager(
            entity_type=EntityManagerTypes.TAGS).get(tag_set_name).tag_set_id

    @property
    def guid(self):
        """Returns the container guid of this Tagset"""
        return self._container_guid

    @property
    def full_name(self):
        """Returns the full name of this Tagset"""
        return self._full_name

    @property
    def comment(self):
        """Returns the comment provided for this Tagset"""
        return self._comment

    @property
    def owner(self):
        """Returns the owner username for this Tagset"""
        return self._owner

    @property
    def owner_alias_name(self):
        """Returns the owner alias name for this Tagset"""
        return self._owner_alias_name

    @property
    def tags(self):
        """Returns the tags present in this tagset"""
        return self._tags

    @property
    def tag_set_id(self):
        """returns the container id for this tagset"""
        return self._tag_set_id


class Tag(object):
    """Class for performing operations on a single Tag"""

    def __init__(self, commcell_object, tag_set_name, tag_name, tag_id=None):
        """Initialize an object of the Tag class.

            Args:
                commcell_object     (object)    --  instance of the commcell class

                tag_set_name     (str)          --  name of the TagSet

                tag_name         (str)          --  Name of tag inside TagSet container

                tag_id       (str)              --  id for tag
                                                        default: None

            Returns:
                object  -   instance of the Tag class
        """
        self._commcell_object = commcell_object
        self._update_response_ = commcell_object._update_response_
        self._services = commcell_object._services
        self._cvpysdk_obj = self._commcell_object._cvpysdk_object
        self._tag_name = tag_name
        self._tag_set_name = tag_set_name
        self._tag_id = None
        self._tag_props = None
        self._api_modify_tag = self._services['GET_TAGS']
        if tag_id is None:
            self._tag_id = self._get_tag_id(tag_set_name, tag_name)
        else:
            self._tag_id = tag_id
        self.refresh()

    def _response_not_success(self, response):
        """Helper function to raise an exception when reponse status is not 200 (OK).

            Args:
                response    (object)    --  response class object,

                received upon running an API request, using the `requests` python package

        """
        raise SDKException('Response', '101', self._update_response_(response.text))

    def _get_tag_id(self, tag_set_name, tag_name):
        """ Get Tag id for given tag name
                Args:

                    tag_set_name    (str)   --  Name of the TagSet

                    tag_name        (str)   --  Name of the Tag

                Returns:

                    int                --  Tag id

        """
        tags = self._commcell_object.activate.entity_manager(
            entity_type=EntityManagerTypes.TAGS)
        # we need this refresh so that tags gets refreshed after adding new tag inside tagset
        tags.refresh()
        tag_set = tags.get(tag_set_name)
        return tag_set.get_tag_id(tag_name=tag_name)

    def _get_tag_properties(self):
        """ Get Tag properties for this associated tag object
                Args:

                    None

                Returns:

                    dict    --  containing tag properties

                        Example : {
                                      "tagOwnerType": 1,
                                      "tagId": 15865,
                                      "name": "p10",
                                      "flags": 0,
                                      "fullName": "cvpysdk1\\p10",
                                      "description": "",
                                      "id": "C9E229D0-B895-4653-9DA7-C9C6BD999121",
                                      "attribute": {}
                                    }

        """

        tags = self._commcell_object.activate.entity_manager(
            entity_type=EntityManagerTypes.TAGS)
        # we need this refresh so that tags gets refreshed after adding new tag inside tagset
        tags.refresh()
        tags.get(self._tag_set_name).refresh()
        tag_set_dict = tags.get_properties(self._tag_set_name)
        tag_dict = tag_set_dict['tagsDetails'][self._tag_name.lower()]
        return tag_dict

    def modify(self, new_name):
        """Modifies the tag name in the tagset

                Args:

                    new_name        (str)       --  New name for Tag

                Returns:

                    None

                Raises:

                    SDKException:

                                if response is empty

                                if response is not success

                                if unable to modify Tag name

                                if input data type is not valid

        """
        if not isinstance(new_name, str):
            raise SDKException('Tags', '101')
        tags = self._commcell_object.activate.entity_manager(
            entity_type=EntityManagerTypes.TAGS)
        tag_set = tags.get(self._tag_set_name)
        request_json = copy.deepcopy(TagConstants.TAG_MODIFY_REQUEST_JSON)
        request_json['container']['containerId'] = tag_set.tag_set_id
        request_json['tags'][0]['tagId'] = self._tag_id
        request_json['tags'][0]['name'] = new_name
        flag, response = self._cvpysdk_obj.make_request(
            'PUT', self._api_modify_tag, request_json
        )
        if flag:
            if response.json():
                if 'errList' in response.json():
                    raise SDKException('Tags', '102', response.json()['errList'][0]['errLogMessage'])
                elif 'tag' in response.json():
                    self._tag_name = new_name
                    self.refresh()
                    return
            raise SDKException('Tags', '105')
        self._response_not_success(response)

    def refresh(self):
        """Refresh the TagSet details for associated object"""
        self._tag_props = self._get_tag_properties()

    @property
    def full_name(self):
        """Returns the full name of the tag inside tagset"""
        return self._tag_props['fullName']

    @property
    def guid(self):
        """Returns the tag guid value"""
        return self._tag_props['id']

    @property
    def tag_id(self):
        """Returns the id of the tag"""
        return self._tag_id


class Classifiers(object):
    """Class for representing all the Classifier entities in the commcell."""

    def __init__(self, commcell_object):
        """Initializes an instance of the Classifiers class.

            Args:
                commcell_object     (object)    --  instance of the commcell class

            Returns:
                object  -   instance of the Classifiers class

        """
        self._commcell_object = commcell_object
        self._update_response_ = commcell_object._update_response_
        self._cvpysdk_object = commcell_object._cvpysdk_object
        self._services = commcell_object._services
        self._classifiers = None
        self._api_create_classifier = self._services['ACTIVATE_ENTITIES']
        self._api_get_classifier = self._services['GET_CLASSIFIERS']
        self._api_delete_classifier = self._services['ACTIVATE_ENTITY']
        self.refresh()

    def _response_not_success(self, response):
        """Helper function to raise an exception when reponse status is not 200 (OK).

            Args:
                response    (object)    --  response class object,

                received upon running an API request, using the `requests` python package

        """
        raise SDKException('Response', '101', self._update_response_(response.text))

    def get_properties(self, classifier_name):
        """Returns a properties of the specified classifier name.

            Args:
                classifier_name (str)  --  name of the classifier

            Returns:
                dict -  properties for the given classifier name


        """
        return self._classifiers[classifier_name.lower()]

    def _get_all_classifier_entities(self):
        """Gets the list of all classifier associated with this commcell.

            Returns:
                dict    -   dictionary consisting of dictionaries, where each dictionary stores the
                                details of a single classifier

            Raises:
                SDKException:
                    if response is empty

                    if response is not success

        """
        flag, response = self._cvpysdk_object.make_request(
            'GET', self._api_get_classifier
        )

        if flag:
            if response.json() and 'entityDetails' in response.json():
                return self._get_classifier_entity_from_collections(response.json())
            raise SDKException('Classifier', '103')
        self._response_not_success(response)

    @staticmethod
    def _get_classifier_entity_from_collections(collections):
        """Extracts all the classifier entities, and their details from the list of collections given,
            and returns the dictionary of all classifier

            Args:
                collections     (list)  --  list of all collections

            Returns:
                dict    -   dictionary consisting of dictionaries, where each dictionary stores the
                                details of a single classifier

        """
        _regex_entity = {}
        for regex_entity in collections['entityDetails']:
            regex_entity_dict = {}
            regex_entity_dict['displayName'] = regex_entity.get('displayName', "")
            regex_entity_dict['entityKey'] = regex_entity.get('entityKey', "")
            regex_entity_dict['categoryName'] = regex_entity.get('categoryName', "")
            regex_entity_dict['entityXML'] = regex_entity.get('entityXML', "")
            regex_entity_dict['entityId'] = regex_entity.get('entityId', 0)
            regex_entity_dict['flags'] = regex_entity.get('flags', 0)
            regex_entity_dict['entityType'] = regex_entity.get('entityType', 0)
            regex_entity_dict['enabled'] = regex_entity.get('enabled', False)
            _regex_entity[regex_entity['entityName'].lower()] = regex_entity_dict
        return _regex_entity

    def refresh(self):
        """Refresh the classifier associated with the commcell."""
        self._classifiers = self._get_all_classifier_entities()

    def delete(self, classifier_name):
        """deletes the specified classifier in the commcell

                    Args:

                        classifier_name (str)      --  name of the classifier

                    Returns:
                        None

                    Raises:

                        SDKException:

                                if response is empty

                                if response is not success

                                if unable to delete classifier in commcell

                                if unable to find classifier in the commcell

                                if data type of input is invalid


                """
        if not isinstance(classifier_name, str):
            raise SDKException('Classifier', '101')
        if classifier_name.lower() not in self._classifiers:
            raise SDKException('Classifier', '102', 'Unable to find given classifier name in the commcell')

        flag, response = self._cvpysdk_object.make_request(
            'DELETE', self._api_delete_classifier % self._classifiers[classifier_name.lower()]['entityId']
        )

        if flag:
            if response.json() and 'errorCode' in response.json() and response.json()['errorCode'] == 0:
                self.refresh()
                return
            raise SDKException('Classifier', '105')
        self._response_not_success(response)

    def add(self, classifier_name, content_analyzer, description="Created from CvPySDK", training_zip_data_file=None):
        """Creates new classifier with given name in the commcell

                Args:

                    classifier_name     (str)       --      Name of the classifier

                    content_analyzer    (str)       --      Content Analyzer cloud name

                    description         (str)       --      Description for classifier

                    training_zip_data_file  (str)   --      Zip file path containing training data files

                Returns:

                    object      --  returns object of Classifier class

                Raises:

                    SDKException:

                        if input data type is not valid

                        if response is empty or not success

                        if failed to create classifier

                        if failed to find content analyzer cloud details

        """

        if not isinstance(classifier_name, str) or not isinstance(content_analyzer, str):
            raise SDKException('Classifier', '101')
        if not self._commcell_object.content_analyzers.has_cloud(content_analyzer):
            raise SDKException('Classifier', '102', "Given CA cloud doesn't exists on this commcell")
        ca_obj = self._commcell_object.content_analyzers.get(content_analyzer)
        request_json = copy.deepcopy(ClassifierConstants.CREATE_REQUEST_JSON)
        request_json['description'] = description
        request_json['entityName'] = classifier_name
        request_json['entityKey'] = classifier_name.replace(" ", "_").lower()
        ca_details_json = request_json['entityXML']['classifierDetails']['CAUsedInTraining']
        ca_details_json['caUrl'] = ca_obj.cloud_url
        ca_details_json['clientId'] = int(ca_obj.client_id)
        ca_details_json['cloudName'] = content_analyzer
        ca_details_json['cloudId'] = int(ca_obj.cloud_id)
        port_no = int(ca_obj.cloud_url.split(":")[2])
        # update if it is not default port no of 22000
        if port_no != 22000:
            request_json['entityXML']['classifierDetails']['trainDatasetURI'] = \
                request_json['entityXML']['classifierDetails']['trainDatasetURI'].replace("22000", str(port_no))
        flag, response = self._cvpysdk_object.make_request(
            'POST', self._api_create_classifier, request_json
        )

        if flag:
            if response.json() and 'entityDetails' in response.json() and 'err' not in response.json():
                entity_id = response.json()['entityDetails'][0]['entityId']
                self.refresh()
                classifier_obj = Classifier(
                    commcell_object=self._commcell_object,
                    classifier_name=classifier_name,
                    entity_id=entity_id)
                if training_zip_data_file:
                    classifier_obj.upload_data(zip_file=training_zip_data_file, start_training=True)
                return classifier_obj
            raise SDKException('Classifier', '104')
        self._response_not_success(response)

    def get(self, classifier_name):
        """Returns a Classifier object for the given classifier name.

            Args:
                classifier_name (str)  --  name of the classifier

            Returns:

                obj                 -- Object of Classifier class

            Raises:
                SDKException:

                    if unable to find classifier info in commcell

                    if classifier_name is not of type string


        """
        if not isinstance(classifier_name, str):
            raise SDKException('Classifier', '101')

        if self.has_classifier(classifier_name.lower()):
            entity_id = self._classifiers[classifier_name.lower()]['entityId']
            return Classifier(self._commcell_object, classifier_name.lower(), entity_id)
        raise SDKException('Classifier', '102', "Unable to get Classifier class object")

    def get_entity_ids(self, classifier_name):
        """Returns a list of entity id for the given classifier name list.

            Args:
                classifier_name (list)  --  names of the classifier

            Returns:

                list                -- entity id's for the given classifier names

            Raises:
                SDKException:

                    if classifier_name is not of type list

                    if unable to find entity id for classifier


        """
        if not isinstance(classifier_name, list):
            raise SDKException('Classifier', '101')
        entity_ids = []
        for classifier in classifier_name:
            classifier = classifier.lower()
            if classifier in self._classifiers:
                entity_ids.append(self._classifiers[classifier]['entityId'])
            else:
                raise SDKException(
                    'Classifier', '102', f"Unable to find entity id for given classifier name :{classifier}")
        return entity_ids

    def get_entity_keys(self, classifier_name):
        """Returns a list of entity keys for the given classifier name list.

            Args:
                classifier_name (list)  --  names of the classifier

            Returns:

                list                -- entity keys for the given classifier names

            Raises:

                SDKException:

                    if classifier_name is not of type list

                    if unable to find entity key for classifier


        """
        if not isinstance(classifier_name, list):
            raise SDKException('Classifier', '101')
        entity_keys = []
        for classifier in classifier_name:
            classifier = classifier.lower()
            if classifier in self._classifiers:
                entity_keys.append(self._classifiers[classifier]['entityKey'])
            else:
                raise SDKException(
                    'Classifier', '102', f"Unable to find entity keys for given classifier name :{classifier}")
        return entity_keys

    def has_classifier(self, classifier_name):
        """Checks if a classifier entity exists in the commcell with the input name.

            Args:
                classifier_name (str)  --  name of the classifier

            Returns:
                bool - boolean output to denote whether classifier exists in the commcell or not

            Raises:
                SDKException:

                    if type of the classifier name argument is not string

        """
        if not isinstance(classifier_name, str):
            raise SDKException('Classifier', '101')

        return self._classifiers and classifier_name.lower() in map(str.lower, self._classifiers)


class Classifier(object):
    """Class for performing operations on a single classifier entity"""

    def __init__(self, commcell_object, classifier_name, entity_id=None):
        """Initialize an object of the Classifier class.

            Args:
                commcell_object     (object)    --  instance of the commcell class

                classifier_name     (str)       --  name of the classifier

                entity_id       (str)           --  id of the classifier
                    default: None

            Returns:
                object  -   instance of the Classifier class
        """
        self._commcell_object = commcell_object
        self._update_response_ = commcell_object._update_response_
        self._services = commcell_object._services
        self._cvpysdk_obj = self._commcell_object._cvpysdk_object
        self._classifier_name = classifier_name
        self._entity_id = None
        self._display_name = None
        self._entity_type = None
        self._is_enabled = None
        self._entity_key = None
        self._entity_xml = None
        self._category_name = None
        self._trained_ca_cloud_id = None
        self._sync_ca_client_id = []
        self._last_training_time = None
        self._training_accuracy = None
        self._training_status = TrainingStatus.NOT_APPLICABLE.value
        self._sample_details = {}
        self._api_upload = self._services['CA_UPLOAD_FILE']
        self._api_start_training = self._services['START_TRAINING']
        self._api_cancel_training = self._services['CANCEL_TRAINING']
        self._api_entity = self._services['ACTIVATE_ENTITY']
        # flags used in upload API call Format : [Version Bit,Message Bit(1-Header, 2-Data Chunk),EOF File Flag]
        self._header_chunk = [1, 1, 0]
        self._data_chunk = [1, 2, 0]
        self._data_chunk_eof = [1, 2, 1]
        if entity_id is None:
            self._entity_id = self._get_entity_id(classifier_name)
        else:
            self._entity_id = entity_id
        self.refresh()

    def _response_not_success(self, response):
        """Helper function to raise an exception when reponse status is not 200 (OK).

            Args:
                response    (object)    --  response class object,

                received upon running an API request, using the `requests` python package

        """
        raise SDKException('Response', '101', self._update_response_(response.text))

    def _get_entity_id(self, classifier_name):
        """ Get entity id for given classifier name
                Args:

                    classifier_name (str)  -- Name of the classifier

                Returns:

                    int                -- entity id of the classifier

        """

        return self._commcell_object.activate.entity_manager(
            EntityManagerTypes.CLASSIFIERS).get(classifier_name.lower()).entity_id

    def _get_upload_api(self):
        """Returns the upload api for this classifier

                Args:
                    None

                Returns:

                    Str     --  Upload API url

        """
        api_params = f"?entityId={self.entity_id}&entityKey={self.entity_key}&waitForCopy=true"
        upload_api = self._api_upload % self.trained_ca_cloud_id
        return f"{upload_api}{api_params}"

    def _get_upload_request_id(self, zip_file):
        """gets the upload request id

                Args:

                    zip_file        (str)       --  Zip file path

                Returns:

                    str     --  Request id for this upload request

                Raises

                    SDKException:

                            if failed to get request id for this upload request

                            if zip file doesn't exists on specified path

                            if it is not a valid zip file

        """
        request_id = None
        if not zip_file.lower().endswith(".zip"):
            raise SDKException('Classifier', '102', "Only Zip files are allowed for model trainings")
        if not os.path.exists(zip_file):
            raise SDKException('Classifier', '107')
        file_stat = os.stat(zip_file)
        xml_req = f"<?xml version='1.0' encoding='UTF-8'?><DM2ContentIndexing_UploadFileReq>" \
                  f"<destEntity clientId=\"0\" _type_=\"3\"/>" \
                  f"<fileContent fileSize=\"{file_stat.st_size}\" fileName=\"{os.path.basename(zip_file)}\"/>" \
                  f"<fileMetaData modifiedTime=\"{str(file_stat.st_mtime).split('.',1)[0]}\" dataType=\"1\"/>" \
                  f"</DM2ContentIndexing_UploadFileReq>"
        req_length = len(xml_req)
        flag_byte = self._get_upload_flag_bit(flags=self._header_chunk, request_data_length=req_length)
        xml_byte = bytearray(xml_req, 'utf-8')
        flag, response = self._cvpysdk_obj.make_request(
            method='POST', url=self._get_upload_api(), payload=flag_byte + xml_byte)
        if flag:
            if response.json():
                if 'errorCode' in response.json():
                    error_code = int(response.json()['errorCode'])

                    if error_code != 0:
                        error_string = response.json()['errorString']
                        raise SDKException(
                            'Classifier', '102', 'Failed to get upload request id : {0}'.format(
                                error_string
                            )
                        )

                if 'requestId' in response.json():
                    request_id = response.json()['requestId']
                return request_id
            raise SDKException('Response', '102')
        raise SDKException('Response', '101', self._update_response_(response.text))

    def _validate_upload_response(self, flag, response, size):
        """Validates the upload response for given chunk size

            Args:

                flag        (bool)      --  flag returned from make_request call for upload API

                response     (resp)     --  response from make_request call for upload API

                size        (int)       --  Chunk size to be checked in response

            Raises:

                SDKException:

                    if response is empty or not success

                    if response size is not required input size

        """

        if flag:
            if response.json():
                if 'errorCode' in response.json():
                    error_code = int(response.json()['errorCode'])

                    if error_code != 0:
                        error_string = response.json()['errorString']
                        raise SDKException(
                            'Classifier', '102', 'Failed to Upload Full file data : {0}'.format(
                                error_string
                            )
                        )

                if 'chunkOffset' in response.json():
                    chunk_offset = response.json()['chunkOffset']
                    if chunk_offset != size:
                        raise SDKException(
                            'Classifier', '102', 'Chunk Offset not matched after upload. Please retry')
                    return
            raise SDKException('Response', '102')
        raise SDKException('Response', '101', self._update_response_(response.text))

    def _get_upload_flag_bit(self, flags, request_data_length):
        """Returns the 7-byte flag used in upload request API

                Args:

                    flags       (list)       --  list containing flag values

                    request_data    (int)   --  length of the request data

                Returns:

                    bytes   --  bytes array representing flags used in upload api call


        """
        # we need to put request length of upload api call into 4 byte flag buffer
        flag_byte = bytearray(flags)
        data_size_bytes = request_data_length.to_bytes(4, byteorder='big')
        flag_byte = flag_byte + data_size_bytes[::-1]
        return flag_byte

    def monitor_training(self, timeout=30):
        """Monitor the training status on this classifier

                Args:

                    timeout     (int)   --  minutes after which the training will not be monitored,

                                                     default: 30

                Returns:

                    bool        --  to denote whether training got completed or not.


        """
        start_time = time.time()
        time_limit = start_time + (timeout * 60)
        training_status = None
        while time_limit > start_time:
            start_time = time.time()
            flag, response = self._cvpysdk_obj.make_request(
                'GET', self._api_entity % self.entity_id)
            if flag:
                if response.json() and 'entityDetails' in response.json() and 'err' not in response.json():
                    entity_details = response.json()['entityDetails'][0]['entityXML']
                    if 'classifierDetails' not in entity_details:
                        raise SDKException('Classifier', '102', "Unable to fetch Classifier training details")
                    training_status = entity_details['classifierDetails']['trainingStatus']
                    if training_status == TrainingStatus.COMPLETED.value:
                        return True
                    elif training_status in [TrainingStatus.NOT_APPLICABLE.value, TrainingStatus.CANCELLED.value,
                                             TrainingStatus.FAILED.value, TrainingStatus.NOT_USABLE.value]:
                        return False
                elif 'err' in response.json() and 'errLogMessage' in response.json()['err']:
                    raise SDKException('Classifier', '102', response.json()['err']['errLogMessage'])
                else:
                    raise SDKException('Classifier', '108')
            else:
                self._response_not_success(response)
            time.sleep(30)
        return False

    def start_training(self, wait_for=True):
        """Starts training on this classifier

                Args:

                    wait_for            (bool)      --  Specifies whether we need to wait or not for training completion

                Returns:

                    None

                Raises:

                    SDKException:

                        if failed to start training

        """
        flag, response = self._cvpysdk_obj.make_request(
            'POST', self._api_start_training % (self.trained_ca_cloud_id, self.entity_id))
        if flag:
            if response.json() and 'entityDetails' in response.json() and 'err' not in response.json():
                if wait_for and not self.monitor_training():
                    raise SDKException('Classifier', '102', "Training ended in Error")
                self.refresh()
                return
            elif 'err' in response.json() and 'errLogMessage' in response.json()['err']:
                raise SDKException('Classifier', '102', response.json()['err']['errLogMessage'])
            raise SDKException('Classifier', '108')
        self._response_not_success(response)

    def cancel_training(self):
        """Cancels training on this classifier

                Args:

                   None

                Returns:

                    None

                Raises:

                    SDKException:

                        if failed to Cancel training

        """
        flag, response = self._cvpysdk_obj.make_request(
            'POST', self._api_cancel_training % (self.trained_ca_cloud_id, self.entity_id))
        if flag:
            if response.json() and 'entityDetails' in response.json() and 'err' not in response.json():
                entity_details = response.json()['entityDetails'][0]['entityXML']
                if 'classifierDetails' not in entity_details:
                    raise SDKException('Classifier', '102', "Unable to fetch Classifier training details")
                training_status = entity_details['classifierDetails'].get('trainingStatus', 0)
                if training_status != TrainingStatus.CANCELLED.value:
                    raise SDKException(
                        'Classifier',
                        '102',
                        f"Wrong training status even after cancelling : {training_status}")
                self.refresh()
                return
            elif 'err' in response.json() and 'errLogMessage' in response.json()['err']:
                raise SDKException('Classifier', '102', response.json()['err']['errLogMessage'])
            raise SDKException('Classifier', '109')
        self._response_not_success(response)

    def modify(self, classifier_new_name=None, description="Modified from CvPySDK", enabled=True):
        """Modifies the classifier entity

                Args:

                    classifier_new_name         (str)       --  New name for classifier

                    description                 (str)       --  Description string for classifier

                    enabled                     (bool)      --  flag to denote whether classifier is enabled or disabled

                Returns:

                    None

                Raises:

                    SDKException:

                            if failed to modify the classifier

        """
        flag, response = self._cvpysdk_obj.make_request(
            'GET', self._api_entity % self.entity_id)
        if flag:
            if response.json() and 'entityDetails' in response.json() and 'err' not in response.json():
                request = copy.deepcopy(response.json()['entityDetails'][0])
                if classifier_new_name:
                    request['entityName'] = classifier_new_name
                    self._classifier_name = classifier_new_name
                if description:
                    request['description'] = description
                if enabled:
                    request['enabled'] = True
                else:
                    request['enabled'] = False
                flag, response = self._cvpysdk_obj.make_request(
                    'PUT', self._api_entity % self.entity_id, request)
                if flag:
                    if response.json() and 'entityDetails' in response.json() and 'err' not in response.json():
                        self.refresh()
                        return
                    elif 'err' in response.json() and 'errLogMessage' in response.json()['err']:
                        raise SDKException('Classifier', '102', response.json()['err']['errLogMessage'])

                    raise SDKException('Classifier', '106')

                self._response_not_success(response)
            elif 'err' in response.json() and 'errLogMessage' in response.json()['err']:
                raise SDKException('Classifier', '102', response.json()['err']['errLogMessage'])

            raise SDKException('Classifier', '106')

        self._response_not_success(response)

    def upload_data(self, zip_file, start_training=False):
        """Uploads the model training data set zip file to content analyzer machine

                Args:

                    zip_file        (str)       --      Zip file path

                    start_training  (bool)      --      Denotes whether to start training on classifier or not

                Returns:

                    None

                Raises

                    SDKException:

                            if failed to upload the file


        """
        chunk_size = 1048576  # 1MB
        request_id = self._get_upload_request_id(zip_file=zip_file)
        file_stat = os.stat(zip_file)
        req_length = len(request_id)
        file_byte = open(zip_file, 'rb')
        flag_byte = self._get_upload_flag_bit(flags=self._data_chunk_eof, request_data_length=req_length)
        xml_byte = bytearray(request_id, 'utf-8')
        if file_stat.st_size <= chunk_size:
            # full file upload
            data_byte = file_byte.read()
            file_byte.close()
            payload = flag_byte + xml_byte + data_byte
            flag, response = self._cvpysdk_obj.make_request(
                method='POST', url=self._get_upload_api(), payload=payload)
            self._validate_upload_response(flag=flag, response=response, size=file_stat.st_size)
        else:
            # chunk based upload
            file_size = file_stat.st_size
            chunk_count = 1
            while file_size > chunk_size:
                flag_byte = self._get_upload_flag_bit(flags=self._data_chunk, request_data_length=req_length)
                file_size = file_size - chunk_size
                data_byte = file_byte.read(chunk_size)
                payload = flag_byte + xml_byte + data_byte
                flag, response = self._cvpysdk_obj.make_request(
                    method='POST', url=self._get_upload_api(), payload=payload)
                self._validate_upload_response(flag=flag, response=response, size=chunk_size * chunk_count)
                chunk_count = chunk_count + 1
            flag_byte = self._get_upload_flag_bit(flags=self._data_chunk_eof, request_data_length=req_length)
            data_byte = file_byte.read(file_size)
            flag, response = self._cvpysdk_obj.make_request(
                method='POST', url=self._get_upload_api(), payload=flag_byte + xml_byte + data_byte)
            self._validate_upload_response(flag=flag, response=response, size=file_stat.st_size)
            file_byte.close()
        if start_training:
            self.start_training(wait_for=True)

    def _get_entity_properties(self):
        """ Get classifier entity properties
                Args:

                    None

                Returns:

                    None

        """
        classifier_obj = self._commcell_object.activate.entity_manager(
            EntityManagerTypes.CLASSIFIERS)
        # Refresh before refetching properties
        classifier_obj.refresh()
        regex_entity_dict = classifier_obj.get_properties(
            self._classifier_name.lower())
        self._display_name = regex_entity_dict['displayName']
        self._category_name = regex_entity_dict['categoryName']
        self._entity_id = regex_entity_dict['entityId']
        self._is_enabled = regex_entity_dict['enabled']
        self._entity_key = regex_entity_dict['entityKey']
        self._entity_type = regex_entity_dict['entityType']
        self._entity_xml = regex_entity_dict['entityXML']
        if 'classifierDetails' in self._entity_xml:
            self._training_status = int(self._entity_xml['classifierDetails'].get('trainingStatus', 0))
            self._training_accuracy = self._entity_xml['classifierDetails'].get('classifierAccuracy', 0)
            self._sample_details['totalSamples'] = self._entity_xml['classifierDetails'].get('totalSamples', 0)
            self._sample_details['trainingSamplesUsed'] = self._entity_xml['classifierDetails'].get(
                'trainingSamplesUsed', 0)
            self._sample_details['validationSamplesUsed'] = self._entity_xml['classifierDetails'].get(
                'validationSamplesUsed', 0)
            if 'CAUsedInTraining' in self._entity_xml['classifierDetails']:
                trained_ca = self._entity_xml['classifierDetails']['CAUsedInTraining']
                # Mandatory trained CA fields. so no need to use dict.get()
                self._trained_ca_cloud_id = int(trained_ca['cloudId'])
                self._last_training_time = trained_ca.get('lastModelTrainTime', 0)
            if 'syncedContentAnalyzers' in self._entity_xml['classifierDetails'] and self._entity_xml[
                    'classifierDetails']['syncedContentAnalyzers'] is not None:
                sync_ca = self._entity_xml['classifierDetails']['syncedContentAnalyzers'].get('contentAnalyzerList', [])
                for content_analyzer in sync_ca:
                    self._sync_ca_client_id.append(content_analyzer['clientId'])
        return regex_entity_dict

    @property
    def entity_id(self):
        """Returns the value of the entity id attribute."""
        return self._entity_id

    @property
    def entity_type(self):
        """Returns the entity type attribute."""
        return self._entity_type

    @property
    def category_name(self):
        """Returns the entity category name attribute."""
        return self._category_name

    @property
    def display_name(self):
        """Returns the entity display name attribute."""
        return self._display_name

    @property
    def is_enabled(self):
        """Returns the entity isenabled attribute."""
        return self._is_enabled

    @property
    def entity_key(self):
        """Returns the entity key attribute."""
        return self._entity_key

    @property
    def entity_xml(self):
        """Returns the entity xml attribute."""
        return self._entity_xml

    @property
    def trained_ca_cloud_id(self):
        """Returns the cloudid attribute for trained CA"""
        return self._trained_ca_cloud_id

    @property
    def training_status(self):
        """Returns the training status attribute"""
        return TrainingStatus(self._training_status).name

    @property
    def training_accuracy(self):
        """Returns the training accuracy attribute"""
        return self._training_accuracy * 100

    @property
    def last_training_time(self):
        """Returns the last training time attribute"""
        return self._last_training_time

    @property
    def sycn_ca_client_id(self):
        """Returns list of Model synced CA client id attribute"""
        return self._sync_ca_client_id

    @property
    def sample_details(self):
        """Returns dict containing model sample count details used for this classifier training"""
        return self._sample_details

    def refresh(self):
        """Refresh the classifier details for associated object"""
        self._get_entity_properties()



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""Main file for performing operations on inventory manager app under Activate.

Inventories, Inventory, Assets & Asset are the four classes defined in this file

Inventories: class for representing all inventories in the commcell

Inventory: class for representing a single inventory in the commcell

Assets: class for representing all assets in an inventory

Asset: class to represent single asset in an inventory

Inventories:

            __init__()                          --  initialise object of the Inventories class

            _get_inventories()                  --  Gets all inventories in the commcell

             _response_not_success()            --  parses through the exception response, and raises SDKException

             refresh()                          --  refresh the Inventories from the commcell

             get_properties()                   --  returns the properties for given inventory name

             has_inventory()                    --  Checks if a given inventory name exists in the commcell or not

             get()                              --  returns the Inventory class object for given inventory name

             add()                              --  add inventory to the commcell

             delete()                           --  delete inventory from the commcell


Inventory:

            __init__()                          --  initialise object of the Inventory class

            _response_not_success()             --  parses through the exception response, and raises SDKException

            _get_inventory_properties()         --  Gets all the properties of this inventory

             _get_schedule_object()             --  returns the schedule class object associated to this inventory

            _get_data_source_handler_object()   --  returns the datasource and default handler object for this inventory

            _get_permission()                   --  returns the security associations for this inventory

            refresh()                           --  refresh the properties of the inventory

            get_assets()                        --  returns the Assets class object for this inventory

            share()                             --  shares inventory with other user or user group

            start_collection()                  --  starts collection job on this inventory

            get_inventory_data()                --  returns data from inventory

Inventory Attributes
-----------------

    **properties**              --  returns properties of the inventory

    **index_server_name**       --  returns the index server name associated with this inventory

    **_index_server_cloud_id**  --  returns the index server cloudid associated with this inventory

    **inventory_name**          --  returns the inventory name

    **inventory_id**            --  returns the inventory id

    **security_associations**   --  returns the security associations blob of this inventory

    **schedule**                --  returns the schedule object associated with this inventory

    **data_source**             --  returns the DataSource object associated with this inventory

    **handler**                 --  returns the default handler object for this inventory

Assets:

        __init__()                          --  initialise object of the Assets class

        refresh()                           --  refresh the assets associated with inventory

        add()                               --  adds asset to the inventory

        get()                               --  returns the instance of Asset class based on given asset name

        has_asset()                         --  returns whether given asset exists or not in inventory

        delete()                            --  deletes the asset from the inventory

        _get_assets_properties()            --  returns the assets properties

         _response_not_success()            --  parses through the exception response, and raises SDKException

Assets Attributes:
----------------

    **assets**                  --  returns the assets details as json

Asset:

        __init__()                          --  initialise object of the Asset class

        _get_properties()                   --  returns the properties of the asset

        _response_not_success()             --  parses through the exception response, and raises SDKException

        refresh()                           --  refresh the asset associated with inventory

        get_job_history()                   --  returns the job history details of this asset

        get_job_status()                    --  returns the latest job status details of this asset

        get_asset_prop()                    --  returns the asset property value for the given property name


Asset Attributes:
-----------------

        **asset_id**            --      returns the id of asset

        **asset_name**          --      returns the name of asset

        **asset_type**          --      returns the type of asset

        **crawl_start_time**    --      returns the last crawl start time of asset

        **asset_props**         --      returns the properties(name/value pair) of asset

        **asset_status**        --      returns the status of asset

        **inventory_id**        --      returns the inventory id of this asset


"""
import copy

from ..activateapps.constants import InventoryConstants
from ..schedules import Schedules
from ..exception import SDKException


class Inventories():
    """Class for representing all inventories in the commcell."""

    def __init__(self, commcell_object):
        """Initializes an instance of the Inventories class.

            Args:
                commcell_object     (object)    --  instance of the commcell class

            Returns:
                object  -   instance of the Inventories class

        """
        self._commcell_object = commcell_object
        self._update_response_ = commcell_object._update_response_
        self._cvpysdk_object = commcell_object._cvpysdk_object
        self._services = commcell_object._services
        self._inventories = None
        self._API_INVENTORIES = self._services['EDISCOVERY_INVENTORIES']
        self._API_DELETE_INVENTORY = self._services['EDISCOVERY_INVENTORY']
        self.refresh()

    def _response_not_success(self, response):
        """Helper function to raise an exception when reponse status is not 200 (OK).

            Args:
                response    (object)    --  response class object,

                received upon running an API request, using the `requests` python package

        """
        raise SDKException('Response', '101', self._update_response_(response.text))

    def _get_inventories(self):
        """Gets all inventories from the commcell

            Args:

                None

            Return:

                list(dict)        --  list Containing inventory details dict

            Raises:

                SDKException:

                    if response is empty

                    if response is not success

        """
        output = {}
        flag, response = self._cvpysdk_object.make_request(
            'GET', self._API_INVENTORIES
        )
        if flag:
            if response.json() and 'inventories' in response.json():
                inventories = response.json()['inventories']
                for inventory in inventories:
                    output[inventory['displayName'].lower()] = inventory
                return output
            raise SDKException('Inventory', '103')
        self._response_not_success(response)

    def add(self, inventory_name, index_server, name_server=None):
        """Adds inventory to the commcell with given inputs

                Args:

                    inventory_name              (str)       --  Name of the inventory

                    index_server                (str)       --  Index server name

                    name_server                 (list)      --  Name server assets which needs to be added to inventory

                Returns:

                    object  --  Instance of Inventory Class

                Raises:

                    SDKException:

                            if input data type is not valid

                            if failed to add inventory

                            if Index Server doesn't exists in commcell

        """
        if not isinstance(inventory_name, str) or not isinstance(index_server, str):
            raise SDKException('Inventory', '101')
        req_json = copy.deepcopy(InventoryConstants.INVENTORY_ADD_REQUEST_JSON)
        if name_server:
            req_json['identityServers'] = name_server
        req_json['name'] = inventory_name
        if not self._commcell_object.index_servers.has(index_server):
            raise SDKException('Inventory', '102', "Given index server name not exists on this commcell")
        index_server_obj = self._commcell_object.index_servers.get(index_server)
        req_json['indexServer']['cloudId'] = index_server_obj.cloud_id
        req_json['indexServer']['displayName'] = index_server_obj.cloud_name
        flag, response = self._cvpysdk_object.make_request(
            'POST', self._API_INVENTORIES, req_json
        )
        if flag:
            if response.json():
                if 'errorCode' in response.json() and response.json()['errorCode'] != 0:
                    raise SDKException(
                        'Inventory',
                        '102',
                        f"Failed to create inventory with error [{response.json()['errorMessage']}]")
                elif 'name' in response.json():
                    inventory = response.json()['name']
                    inventory_id = response.json()['id']
                    self.refresh()
                    return Inventory(self._commcell_object, inventory_name, inventory_id)
                raise SDKException('Inventory', '102', f"Failed to create inventory with response - {response.json()}")
            raise SDKException('Inventory', '105')
        self._response_not_success(response)

    def delete(self, inventory_name):
        """Deletes the inventory from the commcell

                Args:

                    inventory_name      (str)       --  Inventory name to be deleted

                Returns:
                    None

                Raises:

                    SDKException:

                            if unable to find inventory

                            if failed to delete inventory

                            if input type is not valid

        """
        if not isinstance(inventory_name, str):
            raise SDKException('Inventory', '101')
        if not self.has_inventory(inventory_name):
            raise SDKException('Inventory', '106')
        flag, response = self._cvpysdk_object.make_request(
            'DELETE', self._API_DELETE_INVENTORY % self._inventories[inventory_name.lower()]['id']
        )
        if flag:
            if response.json():
                if 'errorCode' in response.json() and response.json()['errorCode'] != 0:
                    raise SDKException(
                        'Inventory',
                        '102',
                        f"Failed to Delete inventory with error [{response.json()['errorMessage']}]")
                self.refresh()
            else:
                raise SDKException('Inventory', '107')
        else:
            self._response_not_success(response)

    def refresh(self):
        """Refresh the inventories associated with the commcell."""
        self._inventories = self._get_inventories()

    def get_properties(self, inventory_name):
        """Returns a properties of the specified Inventory

            Args:
                inventory_name (str)  --  name of the inventory

            Returns:
                dict -  properties for the given inventory name


        """
        return self._inventories[inventory_name.lower()]

    def has_inventory(self, inventory_name):
        """Checks if a inventory exists in the commcell with the input name.

            Args:
                inventory_name (str)  --  name of the inventory

            Returns:
                bool - boolean output to specify whether the inventory exists in the commcell or not

            Raises:
                SDKException:
                    if type of the inventory name argument is not string

        """
        if not isinstance(inventory_name, str):
            raise SDKException('Inventory', '101')

        return self._inventories and inventory_name.lower() in map(str.lower, self._inventories)

    def get(self, inventory_name):
        """Returns a Inventory object for the given inventory name.

            Args:
                inventory_name (str)  --  name of the inventory

            Returns:

                obj                 -- Object of Inventory class

            Raises:

                SDKException:

                    if inventory doesn't exists in commcell

                    if inventory_name is not of type string


        """
        if not isinstance(inventory_name, str):
            raise SDKException('Inventory', '101')

        if self.has_inventory(inventory_name):
            inventory_id = self._inventories[inventory_name.lower()]['id']
            return Inventory(self._commcell_object, inventory_name, inventory_id)
        raise SDKException('Inventory', '106')


class Inventory():
    """Class for performing operations on a single inventory"""

    def __init__(self, commcell_object, inventory_name, inventory_id=None):
        """Initialize an object of the Inventory class.

            Args:
                commcell_object     (object)    --  instance of the commcell class

                inventory_name     (str)        --  name of the Inventory

                inventory_id       (str)        --  id of Inventory
                    default: None

            Returns:
                object  -   instance of the Inventory class
        """
        self._commcell_object = commcell_object
        self._update_response_ = commcell_object._update_response_
        self._services = commcell_object._services
        self._cvpysdk_obj = self._commcell_object._cvpysdk_object
        self._inventory_id = None
        self._inventory_name = inventory_name
        self._inventory_props = None
        self._index_server_name = None
        self._index_server_cloud_id = None
        self._security_associations = None
        self._schedule = None
        self._data_source = None
        self._handler = None
        self._API_GET_INVENTORY_DETAILS = self._services['EDISCOVERY_INVENTORY']
        self._API_SECURITY_ENTITY = self._services['ENTITY_SECURITY_ASSOCIATION']
        self._API_GET_DEFAULT_HANDLER = self._services['EDISCOVERY_GET_DEFAULT_HANDLER']
        self._API_PERMISSION = self._services['V4_ACTIVATE_DS_PERMISSION']
        self._API_CRAWL = self._services['V4_INVENTORY_CRAWL']

        if not inventory_id:
            self._inventory_id = self._commcell_object.activate.inventory_manager().get(inventory_name).inventory_id
        else:
            self._inventory_id = inventory_id
        self.refresh()

    def _response_not_success(self, response):
        """Helper function to raise an exception when reponse status is not 200 (OK).

            Args:
                response    (object)    --  response class object,

                received upon running an API request, using the `requests` python package

        """
        raise SDKException('Response', '101', self._update_response_(response.text))

    def _get_permission(self):
        """returns security association blob for this inventory

                Args:

                    None

                Returns:

                    dict    --  Security association blob

        """
        flag, response = self._cvpysdk_obj.make_request(
            'GET', self._API_PERMISSION % self._inventory_id
        )
        if flag:
            if response.json() and 'securityAssociations' in response.json():
                return response.json()['securityAssociations']
            raise SDKException('Inventory', '102', "Inventory permission fetch failed")
        self._response_not_success(response)

    def _get_inventory_properties(self):
        """ Get inventory properties from the commcell
                Args:

                    None

                Returns:

                    dict        --  Properties of inventory

        """
        flag, response = self._cvpysdk_obj.make_request(
            'GET', self._API_GET_INVENTORY_DETAILS % self._inventory_id
        )
        if flag:
            if response.json():
                inventory_props = response.json()
                self._index_server_name = inventory_props['indexServer']['displayName']
                self._index_server_cloud_id = inventory_props['indexServer']['cloudId']
                self._inventory_name = inventory_props['displayName']
                return inventory_props
            raise SDKException('Inventory', '104')
        self._response_not_success(response)

    def refresh(self):
        """Refresh the inventory details for associated object"""
        self._inventory_props = self._get_inventory_properties()
        self._schedule = self._get_schedule_object()
        self._data_source, self._handler = self._get_data_source_handler_object()
        self._security_associations = self._get_permission()

    def get_assets(self):
        """Returns the Assets class instance for this inventory

                Args:

                    None

                Returns:

                    object      --  Instance of Assets class

        """
        return Assets(self._commcell_object, self.inventory_name, self.inventory_id)

    def start_collection(self):
        """Starts collection job on this inventory

                Args:

                    None

                Return:

                    None

                Raises:

                    SDKException:

                            if failed to start collection job

        """
        flag, response = self._cvpysdk_obj.make_request(
            'PUT', self._API_CRAWL % self._inventory_id
        )
        if flag:
            if response.json():
                if 'errorCode' in response.json() and response.json()['errorCode'] != 0:
                    raise SDKException(
                        'Inventory',
                        '102',
                        f"Failed to start crawl on inventory with error [{response.json()['errorMessage']}]")
                return
            else:
                raise SDKException('Inventory', '102', "Unknown response while starting collection job on inventory")
        else:
            self._response_not_success(response)

    def share(self, user_or_group_name, allow_edit_permission=False, is_user=True, ops_type=1):
        """Shares inventory with given user or user group in commcell

                Args:

                    user_or_group_name      (str)       --  Name of user or group

                    is_user                 (bool)      --  Denotes whether this is user or group name
                                                                default : True(User)

                    allow_edit_permission   (bool)      --  whether to give edit permission or not to user or group

                    ops_type                (int)       --  Operation type

                                                            Default : 1 (Add)

                                                            Supported : 1 (Add)
                                                                        3 (Delete)

                Returns:

                    None

                Raises:

                    SDKException:

                            if unable to update security associations

                            if response is empty or not success
        """
        if not isinstance(user_or_group_name, str):
            raise SDKException('Inventory', '101')
        request_json = copy.deepcopy(InventoryConstants.INVENTORY_SHARE_REQUEST_JSON)
        external_user = False
        association_response = None
        if ops_type == 1 and len(self.security_associations) > 1:
            association_request_json = copy.deepcopy(InventoryConstants.INVENTORY_SHARE_REQUEST_JSON)
            del association_request_json['securityAssociations']
            association_request_json['entityAssociated']['entity'][0]['seaDataSourceId'] = int(self.inventory_id)
            # get security blob for this data source type entity - 132
            flag, response = self._cvpysdk_obj.make_request(
                'GET', self._API_SECURITY_ENTITY % (132, int(self.inventory_id)), association_request_json
            )
            if flag:
                if response.json() and 'securityAssociations' in response.json():
                    association_response = response.json(
                    )['securityAssociations'][0]['securityAssociations']['associations']
                else:
                    raise SDKException('Inventory', '102', 'Failed to get existing security associations')
            else:
                response_string = self._commcell_object._update_response_(response.text)
                raise SDKException('Response', '101', response_string)

        if '\\' in user_or_group_name:
            external_user = True
        if is_user:
            user_obj = self._commcell_object.users.get(user_or_group_name)
            user_id = user_obj.user_id
            request_json['securityAssociations']['associations'][0]['userOrGroup'][0]['userId'] = int(user_id)
            request_json['securityAssociations']['associations'][0]['userOrGroup'][0]['_type_'] = 13
            request_json['securityAssociations']['associations'][0]['userOrGroup'][0]['userName'] = user_or_group_name
        elif external_user:
            request_json['securityAssociations']['associations'][0]['userOrGroup'][0]['groupId'] = 0
            request_json['securityAssociations']['associations'][0]['userOrGroup'][0]['_type_'] = 62
            request_json['securityAssociations']['associations'][0]['userOrGroup'][0][
                'externalGroupName'] = user_or_group_name
        else:
            grp_obj = self._commcell_object.user_groups.get(user_or_group_name)
            grp_id = grp_obj.user_group_id
            request_json['securityAssociations']['associations'][0]['userOrGroup'][0]['userGroupId'] = int(grp_id)
            request_json['securityAssociations']['associations'][0]['userOrGroup'][0]['_type_'] = 15
            request_json['securityAssociations']['associations'][0]['userOrGroup'][0][
                'userGroupName'] = user_or_group_name

        request_json['entityAssociated']['entity'][0]['seaDataSourceId'] = self.inventory_id
        request_json['securityAssociations']['associationsOperationType'] = ops_type

        if allow_edit_permission:
            request_json['securityAssociations']['associations'][0]['properties']['categoryPermission']['categoriesPermissionList'].append(
                InventoryConstants.EDIT_CATEGORY_PERMISSION)

            # Associate existing associations to the request
        if ops_type == 1 and len(self.security_associations) > 1:
            request_json['securityAssociations']['associations'].extend(association_response)

        flag, response = self._cvpysdk_obj.make_request(
            'PUT', self._API_PERMISSION % self._inventory_id, request_json
        )
        if flag:
            if response.json():
                error_code = response.json()['errorCode']
                if error_code != 0:
                    error_message = response.json()['errorString']
                    raise SDKException(
                        'Inventory',
                        '102', error_message)
                # update association list by refreshing inventory
                self.refresh()
            else:
                raise SDKException('Inventory', '111')
        else:
            response_string = self._commcell_object._update_response_(response.text)
            raise SDKException('Response', '101', response_string)

    def _get_data_source_handler_object(self):
        """returns the data source and handler object associated with this inventory

                Args:
                    None

                Returns:
                    obj,obj     --  Instance of DataSource object,Instance of Handler object

                Raises:

                    SDKException:

                            if failed to get datasource or handler details
        """
        flag, response = self._cvpysdk_obj.make_request(
            'GET', self._API_GET_DEFAULT_HANDLER % self.inventory_id)
        if flag:
            if response.json() and 'handlerInfos' in response.json():
                handler_list = response.json()['handlerInfos']
                if not isinstance(handler_list, list):
                    raise SDKException('Inventory', '102', "Failed to get Datasource/Handler details")
                handler_details = handler_list[0]
                ds_name = handler_details['dataSourceName']
                handler_name = handler_details['handlerName']
                ds_obj = self._commcell_object.datacube.datasources.get(ds_name)
                handler_obj = ds_obj.ds_handlers.get(handler_name)
                return ds_obj, handler_obj
            raise SDKException('Inventory', '102', 'Unknown response while fetching datasource details')
        response_string = self._commcell_object._update_response_(response.text)
        raise SDKException('Response', '101', response_string)

    def _get_schedule_object(self):
        """Returns the schedule class object of schedule associated to this inventory

                Args:

                    None

                Returns:

                    object      --  Instance of Schedule class

                Raises:

                    SDKException:

                            if failed to find schedule
        """
        return Schedules(self).get()

    def get_inventory_data(self, handler_filter=""):
        """ Executes handler for fetching data from inventory
                Args:

                     handler_filter    (str)  -- Filter which needs to applied for handler execution

                Returns:

                    dict        --  Dictionary of values fetched from handler execution

                Raises:

                    SDKExpception:

                        if error in fetching handler data

                        if input is not valid
        """
        if not isinstance(handler_filter, str):
            raise SDKException('Inventory', '101')
        if not self._handler:
            raise SDKException('Inventory', '102', 'No handler object initialised')
        return self._handler.get_handler_data(handler_filter=handler_filter)

    @property
    def properties(self):
        """Returns the properties of this inventory as dict"""
        return self._inventory_props

    @property
    def index_server_name(self):
        """Returns the index server name associated with this inventory"""
        return self._index_server_name

    @property
    def index_server_cloud_id(self):
        """Returns the index server id associated with this inventory"""
        return self._index_server_cloud_id

    @property
    def inventory_id(self):
        """Returns the inventory id associated with this inventory"""
        return self._inventory_id

    @property
    def inventory_name(self):
        """Returns the inventory name associated with this inventory"""
        return self._inventory_name

    @property
    def security_associations(self):
        """Returns the security blob associated with this inventory"""
        return self._security_associations

    @property
    def schedule(self):
        """Returns the schedule class object for schedule associated with this inventory"""
        return self._schedule

    @property
    def data_source(self):
        """Returns the DataSource class object for datasource associated with this inventory"""
        return self._data_source

    @property
    def handler(self):
        """Returns the Handler class object for default handler associated with this inventory"""
        return self._handler


class Assets():
    """Class to represent all assets in an inventory"""

    def __init__(self, commcell_object, inventory_name, inventory_id=None):
        """Initialize an object of the Assets class.

            Args:
                commcell_object     (object)    --  instance of the commcell class

                inventory_name     (str)        --  name of the Inventory

                inventory_id       (str)        --  id of Inventory
                    default: None

            Returns:
                object  -   instance of the Assets class
        """
        self._commcell_object = commcell_object
        self._update_response_ = commcell_object._update_response_
        self._services = commcell_object._services
        self._cvpysdk_obj = self._commcell_object._cvpysdk_object
        self._inventory_id = None
        self._inventory_name = inventory_name
        if not inventory_id:
            self._inventory_id = self._commcell_object.activate.inventory_manager().get(inventory_name).inventory_id
        else:
            self._inventory_id = inventory_id
        self._assets = None
        self._API_ASSETS = self._services['EDISCOVERY_ASSETS']
        self._API_ASSET = self._services['EDISCOVERY_ASSET']
        self.refresh()

    def _get_assets_properties(self):
        """gets the assets properties from inventory

                    Args:
                        None

                    Returns:

                        dict    --  containing asset properties

        """
        flag, response = self._cvpysdk_obj.make_request(
            'GET', self._API_ASSETS % self._inventory_id)
        if flag:
            if response.json() and 'assets' in response.json():
                assets = {}
                for asset in response.json()['assets']:
                    name = asset['name'].lower()
                    assets[name] = asset
                return assets
            else:
                raise SDKException('Inventory', '102', "Unknown response while fetching assets on inventory")
        else:
            self._response_not_success(response)

    def refresh(self):
        """Refresh the assets details associated with this inventory"""
        self._assets = self._get_assets_properties()

    def _response_not_success(self, response):
        """Helper function to raise an exception when reponse status is not 200 (OK).

            Args:
                response    (object)    --  response class object,

                received upon running an API request, using the `requests` python package

        """
        raise SDKException('Response', '101', self._update_response_(response.text))

    def has_asset(self, asset_name):
        """Checks whether given asset exists in inventory or not

                        Args:

                            asset_name      (str)       --  Name of the asset

                        Returns:

                            bool      --  true if asset exists else false
        """
        return self._assets and asset_name.lower() in self._assets

    def get(self, asset_name):
        """Returns the asset object

                Args:

                    asset_name      (str)       --  Name of the asset

                Returns:

                    object      -- Instance of Asset class

                Raises:

                    SDKException:

                            if input is not valid

                            if asset doesn't exists in inventory
        """
        if not isinstance(asset_name, str):
            raise SDKException('Inventory', '101')
        if not self.has_asset(asset_name):
            raise SDKException('Inventory', '109')
        return Asset(self._commcell_object, self._inventory_name, asset_name, self._inventory_id)

    def add(self, asset_name, asset_type=InventoryConstants.AssetType.IDENTITY_SERVER, **kwargs):
        """Adds asset to the inventory

                Args:

                    asset_name          (str)       --  Name of the asset

                    asset_type          (Enum)      --  type of asset (Refer to InventoryConstants.AssetType class)

                    kwargs for FILE SERVER type Asset:

                            fqdn                --  File server FQDN

                            os                  --  File Server OS type
                                                        (Default:Windows)

                            ip                  --  File server IP

                            country_code        --  Country code (ISO 3166 2-letter code)

                Returns:

                    object  --  Instance of Asset class

                Raises:

                    SDKException:

                            if input is not valid

                            if failed to add asset to inventory

        """
        if not isinstance(asset_name, str) or not isinstance(asset_type, InventoryConstants.AssetType):
            raise SDKException('Inventory', '101')
        request_json = {}
        if asset_type.value == InventoryConstants.AssetType.FILE_SERVER.value:
            for prop in InventoryConstants.ASSET_FILE_SERVER_PROPERTY:
                prop_name = InventoryConstants.FIELD_PROPS_MAPPING[prop]
                default_value = ""
                if prop_name not in kwargs:
                    # always use asset name as file server name
                    if prop == InventoryConstants.FIELD_PROPERTY_NAME:
                        default_value = asset_name
                    if prop == InventoryConstants.FIELD_PROPERTY_OS:
                        default_value = "Windows"
                request_json[prop] = kwargs.get(prop_name, default_value)
        else:
            request_json = copy.deepcopy(InventoryConstants.IDENTITY_SERVER_ASSET_ADD_TO_INVENTORY_JSON)
            request_json['identityServers'] = [asset_name]

        flag, response = self._cvpysdk_obj.make_request(
            'PUT', self._API_ASSETS % self._inventory_id, request_json)
        if flag:
            if response.json():
                if 'errorCode' in response.json() and response.json()['errorCode'] != 0:
                    raise SDKException(
                        'Inventory',
                        '102',
                        f"Failed to add asset to inventory with error [{response.json()['errorMessage']}]")
                self.refresh()
                return Asset(self._commcell_object, self._inventory_name, asset_name, self._inventory_id)
            raise SDKException('Inventory', '108')
        self._response_not_success(response)

    def delete(self, asset_name):
        """Delete the asset from the inventory

                Args:

                    asset_name      (str)       --  Name of the asset

                Returns:

                    None

                Raises:

                    SDKException:

                            if input is not valid

                            if failed to delete the asset

                            if unable to find this asset in inventory
        """

        if not isinstance(asset_name, str):
            raise SDKException('Inventory', '101')
        if not self.has_asset(asset_name):
            raise SDKException('Inventory', '109')
        flag, response = self._cvpysdk_obj.make_request(
            'DELETE', self._API_ASSET % (self._inventory_id, self._assets[asset_name.lower()]['id']))
        if flag:
            if response.json():
                if 'errorCode' in response.json() and response.json()['errorCode'] != 0:
                    raise SDKException(
                        'Inventory',
                        '102',
                        f"Failed to delete asset from inventory with error [{response.json()['errorMessage']}]")
                self.refresh()
                return
            raise SDKException('Inventory', '110')
        self._response_not_success(response)

    @property
    def assets(self):
        """Returns the assets details associated with this inventory"""
        return self._assets


class Asset():
    """Class to represent single asset in an inventory"""

    def __init__(self, commcell_object, inventory_name, asset_name, inventory_id=None):
        """Initialize an object of the Asset class.

            Args:
                commcell_object     (object)    --  instance of the commcell class

                inventory_name     (str)        --  name of the Inventory

                asset_name          (str)       --  Name of the asset

                inventory_id       (str)        --  id of Inventory
                    default: None

            Returns:
                object  -   instance of the Asset class
        """
        self._commcell_object = commcell_object
        self._update_response_ = commcell_object._update_response_
        self._services = commcell_object._services
        self._cvpysdk_obj = self._commcell_object._cvpysdk_object
        self._inventory_id = None
        self._inventory_name = inventory_name
        if not inventory_id:
            self._inventory_id = self._commcell_object.activate.inventory_manager().get(inventory_name).inventory_id
        else:
            self._inventory_id = inventory_id
        self._asset_name = asset_name
        self._asset_props = None
        self._asset_id = None
        self._crawl_start_time = None
        self._asset_type = None
        self._asset_status = None
        self._asset_name_values_props = None
        self._API_ASSETS = self._services['EDISCOVERY_ASSETS']
        self._API_ASSET = self._services['EDISCOVERY_ASSET']
        self._API_ASSET_JOBS = self._services['EDISCOVERY_ASSET_JOBS']
        self.refresh()

    def _response_not_success(self, response):
        """Helper function to raise an exception when reponse status is not 200 (OK).

            Args:
                response    (object)    --  response class object,

                received upon running an API request, using the `requests` python package

        """
        raise SDKException('Response', '101', self._update_response_(response.text))

    def _get_properties(self):
        """Returns the properties of this asset

                Args:

                    None

                Returns:

                    dict        -- Containing properties of asset

        """

        flag, response = self._cvpysdk_obj.make_request(
            'GET', self._API_ASSETS % self._inventory_id)
        if flag:
            if response.json() and 'assets' in response.json():
                for asset in response.json()['assets']:
                    if asset['name'].lower() == self._asset_name.lower():
                        self._asset_id = asset.get('id', "")
                        self._crawl_start_time = asset.get('lastCollectionTime', 0)
                        self._asset_type = asset.get('type', "")
                        self._asset_status = asset.get('status', "NA")
                        self._asset_name = asset['name']
                        if self._asset_type == InventoryConstants.AssetType.FILE_SERVER.name:
                            flag, response = self._cvpysdk_obj.make_request(
                                'GET', self._API_ASSET % (self._inventory_id, self._asset_id))
                            if flag:
                                if response.json() and 'properties' in response.json():
                                    self._asset_name_values_props = response.json()['properties']
                        return asset
                return {}
            else:
                raise SDKException('Inventory', '102', "Unknown response while fetching assets on inventory")
        else:
            self._response_not_success(response)

    def refresh(self):
        """Refresh the asset details associated with this"""
        self._asset_props = self._get_properties()

    def get_job_history(self):
        """Returns the job history details of this asset

                Args:
                    None

                Returns:

                    list(dict)    --  containing job history details

                Raises:

                    SDKException:

                            if failed to get job history

                            if asset is not supported for this operation

        """
        if self.asset_type != InventoryConstants.AssetType.IDENTITY_SERVER.name:
            raise SDKException('Inventory', '102', "Not supported other than Name Server asset type")
        flag, response = self._cvpysdk_obj.make_request(
            'GET', self._API_ASSET_JOBS % (self._inventory_id, self._asset_id))
        if flag:
            if response.json() and 'jobs' in response.json():
                return response.json()['jobs']
            raise SDKException('Inventory', '102', "Unknown response while fetching job history on inventory")
        else:
            self._response_not_success(response)

    def get_job_status(self):
        """Returns the latest job status details of this asset

                Args:
                    None

                Returns:

                    str    --  last job status (Eg:- RUNNING / IDLE)

                Raises:

                    SDKException:

                            if failed to get job status

                             if asset is not supported for this operation

        """
        if self.asset_type != InventoryConstants.AssetType.IDENTITY_SERVER.name:
            raise SDKException('Inventory', '102', "Not supported other than Name Server asset type")
        self.refresh()  # do refresh before getting current status
        return self.asset_status

    def get_asset_prop(self, prop_name):
        """returns the property value for given property name for this asset

            Args:

                prop_name       (str)       --  Name of the property

            Returns:

                str     --  Value of the property


        """
        for prop in self._asset_name_values_props:
            name = prop['name']
            if name == prop_name:
                return prop['value']
        return ""

    @property
    def asset_id(self):
        """Returns the asset id with this asset"""
        return self._asset_id

    @property
    def asset_name(self):
        """Returns the asset name with this asset"""
        return self._asset_name

    @property
    def crawl_start_time(self):
        """Returns the crawl start time with this asset"""
        return self._crawl_start_time

    @property
    def asset_type(self):
        """Returns the asset type for this asset"""
        return self._asset_type

    @property
    def asset_status(self):
        """Returns the asset status for this asset"""
        return self._asset_status

    @property
    def asset_props(self):
        """Returns the property values for this asset"""
        return self._asset_name_values_props

    @property
    def inventory_id(self):
        """Returns the inventory id for this asset"""
        return self._inventory_id



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""
OneDriveClient class is defined in this file.

OneDriveClient:     Class for a single OneDrive for Business client (v2) of the commcell

OneDriveClient
=======

_get_subclient() --  Returns the sub-client object for OneDrive for Business client (v2)

backup_all_users_in_client() -- Run backup for all users present in OneDrive for Business client (v2)

in_place_restore()  --  Run an inplace restore of specified users for OneDrive for business client (v2)

out_of_place_restore()  --  Run an out-of-place restore of specified users for OneDrive for business client (v2)

disk_restore()  --  Runs disk restore of specified users for OneDrive for business client (v2)
"""

from ..client import Client


class OneDriveClient(Client):
    def __init__(self, commcell_object, client_name, client_id=None):
        """Initialise the OneDrive Client class instance.

            Args:
                commcell_object (object)     --  instance of the Commcell class

                client_name     (str)        --  name of the client

                client_id       (str)        --  id of the client
                                                default: None

            Returns:
                object - instance of the OneDrive Client class
        """
        super(OneDriveClient, self).__init__(commcell_object, client_name, client_id)

    def _get_subclient(self):
        """ Returns the sub-client object for OneDrive for Business client

            Returns:
                _subclient (object) -   Subclient object

        """
        _client = self._commcell_object.clients.get(self.client_name)
        _agent = _client.agents.get('Cloud Apps')
        _instance = _agent.instances.get('OneDrive')
        _backupset = _instance.backupsets.get('defaultbackupset')
        _subclient = _backupset.subclients.get('default')
        return _subclient

    def backup_all_users_in_client(self):
        """ Run backup for all users present in OneDrive client

            Returns:
                object - instance of the Job class for this backup job
        """
        _subclient_object = self._get_subclient()
        return _subclient_object.backup(backup_level='INCREMENTAL')

    def in_place_restore(self, users, **kwargs):
        """ Run an inplace restore of specified users for OneDrive for business client

            Args:
                users (list) :  List of SMTP addresses of users
                **kwargs (dict) : Additional parameters
                    overwrite (bool) : unconditional overwrite files during restore (default: False)
                    restore_as_copy (bool) : restore files as copy during restore (default: False)
                    skip_file_permissions (bool) : If True, restore of file permissions are skipped (default: False)

            Returns:
                object - instance of the Job class for this restore job

            Raises:
                SDKException:
                    if inputs are not of correct type as per definition

                    if failed to initialize job

                    if response is empty

                    if response is not success

        """
        _subclient_object = self._get_subclient()
        restore_job = _subclient_object.in_place_restore_v2(users, **kwargs)
        return restore_job

    def out_of_place_restore(self, users, destination_path, **kwargs):
        """ Run an out-of-place restore of specified users for OneDrive for business client

            Args:
                users (list) : list of SMTP addresses of users
                destination_path (str) : SMTP address of destination user
                **kwargs (dict) : Additional parameters
                    overwrite (bool) : unconditional overwrite files during restore (default: False)
                    restore_as_copy (bool) : restore files as copy during restore (default: False)
                    skip_file_permissions (bool) : If True, restore of file permissions are skipped (default: False)

            Returns:
                object - instance of the Job class for this restore job

            Raises:
                SDKException:
                    if inputs are not of correct type as per definition

                    if failed to initialize job

                    if response is empty

                    if response is not success

        """
        _subclient_object = self._get_subclient()
        restore_job = _subclient_object.out_of_place_restore_v2(users, destination_path, **kwargs)
        return restore_job

    def disk_restore(self,
                     users,
                     destination_client,
                     destination_path,
                     skip_file_permissions=False):
        """ Runs disk restore of specified users for OneDrive for business client

               Args:
                users (list) : list of SMTP addresses of users
                destination_client (str) : client where the users need to be restored
                destination_path (str) : Destination folder location
                skip_file_permissions (bool) : If True, restore of file permissions are skipped (default: False)

            Returns:
                object - instance of the Job class for this restore job

            Raises:
                SDKException:
                    if inputs are not of correct type as per definition

                    if failed to initialize job

                    if response is empty

                    if response is not success

        """
        _subclient_object = self._get_subclient()
        restore_job = _subclient_object.disk_restore_v2(users,
                                                        destination_client,
                                                        destination_path,
                                                        skip_file_permissions=skip_file_permissions)
        return restore_job



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""File for operating on a Virtual Server OracleCloud Subclient.

OracleCloudeVirtualServerSubclient is the only class defined in this file.

OracleCloudVirtualServerSubclient: Derived class from VirtualServerSubClient  Base class,
                            representing a OracleCloud Subclient, and
                            to perform operations on that Subclient

OracleCloudVirtualServerSubclient:

    full_vm_restore_out_of_place()  --  restores the VM specified in to the specified client,
                                        at the specified destination location

"""

from ..vssubclient import VirtualServerSubclient


class OracleCloudVirtualServerSubclient(VirtualServerSubclient):
    """Derived class from VirtualServerSubclient Base class.
       This represents a OracleCloud virtual server subclient,
       and can perform restore operations on only that subclient.

    """

    def full_vm_restore_out_of_place(
            self,
            vm_to_restore=None,
            destination_client=None,
            proxy_client=None,
            new_name=None,
            host=None,
            power_on=True,
            copy_precedence=0,
            restore_option=None,
            **kwargs):
        """Restores the FULL Virtual machine specified in the input list
            to the provided vcenter client along with the ESX and the datastores.
            If the provided client name is none then it restores the Full Virtual
            Machine to the source client and corresponding ESX and datastore.

            Args:
                vm_to_restore           (list)  --  list of all VMs to restore

                destination_client      (str)   --  name of the pseudo client where VM should be
                                                        restored

                proxy_client            (str)   --  the proxy to be used for restore

                new_name                (str)   --  new name to be given to the restored VM

                host                    (str)   --  destination host or cluster; restores to the
                                                        source VM ESX if this value is not
                                                        specified

                power_on                (bool)  --  power on the restored VM
                                                        default: True

                copy_precedence         (int)   --  copy precedence to restored from
                                                        default: 0

                restore_option          (dict)  --  dictionary with all the advanced restore
                                                        options.

                **kwargs                         : Arbitrary keyword arguments Properties as of
                                                     full_vm_restore_out_of_place
                    eg:
                    v2_details          (dict)       -- details for v2 subclient
                                                    eg: check clients.vmclient.VMClient._child_job_subclient_details

            Returns:
                object - instance of the Job class for this restore job

            Raises:
                SDKException:
                    if inputs are not of correct type as per definition

                    if failed to initialize job

                    if response is empty

                    if response is not success

        """
        if not restore_option:
            restore_option = {}
        restore_option["v2_details"] = kwargs.get("v2_details", None)

        # set attr for all the option in restore xml from user inputs
        self._set_restore_inputs(
            restore_option,
            vm_to_restore=self._set_vm_to_restore(vm_to_restore),
            power_on=power_on,
            copy_preceedence=copy_precedence,
            volume_level_restore=1,
            client_name=proxy_client,
            vcenter_client=destination_client,
            esx_host=host,
            out_place=True,
            restore_new_name=new_name
        )

        request_json = self._prepare_fullvm_restore_json(restore_option)
        return self._process_restore_response(request_json)



# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""File for performing index pool related operations on the commcell

IndexPools and IndexPool are 2 classes defined in this file

IndexPools:   Class for representing all the index pools associated with the commcell

IndexPool:    Class for a instance of a single index pool of the commcell


IndexPools
============

    __init__()                          --  initialize object of IndexPools class associated with
    the commcell

    __str()                             --  returns all the Index pools of the commcell

    __repr__()                          --  returns the string to represent the instance

    __getitem__()                       --  returns the details of index pool for given pool name

    _get_all_index_pools()              --  gets detail of all index pools using REST API call

    _response_not_success()             --  raise exception when response is not 200

    get()                               --  return an IndexPool object for given pool name

    has_pool()                          --  returns whether the index pool is present or not in
    the commcell

    add()                            --  creates a new Index pool to the commcell

    delete()                            --  deletes the index pool associated to commcell

    refresh()                           --  refresh the index pools details associated with commcell

IndexPools Attributes
-----------------------

    **all_index_pools**                 --  returns the dictionary consisting of all the index
    pools associated with the commcell and there details


IndexPool
=========

    __init__()                          --  initialize object of IndexPool class

    __repr__()                          --  returns the string to represent the instance

    _response_not_success()             --  raise exception when response is not 200

    modify_node()                       --  modify/add a node to the index pool

    node_info()                         --  returns a dict consisting details of node present in the pool

    refresh()                           --  refresh the index pool details associated with commcell

IndexPool Attributes
----------------------

    **pool_id**                         --  returns the pseudo client id for index pool cloud

    **cloud_id**                        --  returns the cloud id for index pool

    **node_names**                      --  returns a list of names of all the nodes present in pool

    **hac_cluster**                     --  returns the hac cluster name

    **pool_name**                       --  returns the client name for index pool

"""

from copy import deepcopy
from .exception import SDKException
from .datacube.constants import IndexServerConstants


class IndexPools(object):
    """Class for representing all the Index pools associated with the commcell"""

    def __init__(self, commcell_object):
        """Initialize object of the IndexPools class

            Args:
                commcell_object     (object)    --  instance of the Commcell class

            Returns:
                object  -   instance of IndexPools class
        """
        self._commcell_object = commcell_object
        self._cvpysdk_object = commcell_object._cvpysdk_object
        self._services = commcell_object._services
        self._update_response_ = commcell_object._update_response_
        self._all_index_pools = None
        self.refresh()

    def __str__(self):
        """Representation string consisting of all Index pools of the commcell.

                Returns:
                    str - string of all the index pools associated with the commcell
        """
        representation_string = '{:^5}\t{:^20}\n\n'.format('S. No.', 'Index pool Name')
        index = 1
        for pool_name in self.all_index_pools:
            representation_string += '{:^5}\t{:^20}\n'.format(
                index, pool_name)
            index += 1
        return representation_string

    def __repr__(self):
        """Representation string for the instance of the IndexPools class."""
        return "IndexPools class instance for Commcell: '{0}'".format(
            self._commcell_object.commserv_name
        )

    def __getitem__(self, value):
        """Returns the details of index pool for given index pool name

            Args:
                value   (str)       --  name of index pool

            Returns:
                dict    -   details of the index pool

            Raises:
                Index pool not found
        """
        value = value.lower()
        if value in self.all_index_pools:
            return {"pool_name": value, "pool_id": self.all_index_pools[value]}
        raise SDKException('IndexPools', '102')

    def _get_all_index_pools(self):
        """Method to get details of all the index pools present in commcell

            Raises:
                SDKException:
                    Response was not success

                    Response was empty
        """
        flag, response = self._cvpysdk_object.make_request('GET', self._services['GET_ALL_CLIENTS'])
        if flag:
            if response.json() and 'clientProperties' in response.json():
                for dictionary in response.json()['clientProperties']:
                    if dictionary['clientProps']['clusterType'] == 14:
                        temp_name = dictionary['client']['clientEntity']['clientName'].lower()
                        temp_id = int(dictionary['client']['clientEntity']['clientId'])
                        self._all_index_pools[temp_name] = temp_id
            else:
                raise SDKException('Response', '102')
        else:
            raise SDKException('Response', '101', self._update_response_(response.text))

    def _response_not_success(self, response):
        """Helper method to raise exception when response is not 200 (ok)
            Args:
                response    (object)    -   response object

            Raises:
                SDKException:
                    Response was not success
        """
        raise SDKException(
            'Response',
            '101',
            self._update_response_(
                response.text))

    def refresh(self):
        """Refreshes the properties of IndexPools class"""
        self._commcell_object.clients.refresh()
        self._all_index_pools = {}
        self._get_all_index_pools()

    @property
    def all_index_pools(self):
        """Returns a dict consisting details of all index pools

            Returns:
                dict    -   dictionary consisting details for all index pools
                Sample:
                    {
                        <pool_name_1> : <pool_id_1>,
                        <pool_name_2> : <pool_id_2>
                    }
        """
        return self._all_index_pools

    def get(self, pool_name):
        """Returns the IndexPool class object with given pool_name

        Args:
            pool_name       (int/str)       --  Index pool name present in commcell

        Returns:
            object  -   instance of IndexPool class

        Raises:
            SDKExecption:
                Data type of the input(s) is not valid

                Index pool not found
        """
        if isinstance(pool_name, int):
            for index_pool_name in self.all_index_pools:
                if self.all_index_pools[index_pool_name] == pool_name:
                    return IndexPool(self._commcell_object, index_pool_name)
        elif isinstance(pool_name, str):
            if pool_name.lower() in self.all_index_pools:
                return IndexPool(self._commcell_object, pool_name.lower())
        else:
            raise SDKException('IndexPools', '101')
        raise SDKException('IndexPools', '102')

    def has_pool(self, pool_name):
        """Returns whether a index pool is present in the commcell or not

        Args:
            pool_name       (str)       --  Index pool name

        Returns:
            bool    -   True if index pool with given pool_name is present else False
        """
        return pool_name.lower() in self.all_index_pools

    def add(self, pool_name, node_names, hac_name):
        """Creates a new Index pool within the commcell

        Args:
            pool_name       (str)   --  Name for the index pool
            node_names      (list)  --  List of strings of all the index pool node names
            hac_name        (str)   --  Name of the HAC cluster to be used while creating pool

        Raises:
            SDKExecption:
                Data type of the input(s) is not valid.

                Response was not success.

                Response was empty.

        Returns:
            object  -   Returns a object of class IndexPool

        """
        if not (isinstance(pool_name, str) and isinstance(node_names, list)
                and isinstance(hac_name, str)):
            raise SDKException('IndexPools', '101')
        req_json = deepcopy(IndexServerConstants.REQUEST_JSON)
        del req_json['solrCloudInfo']
        del req_json['cloudMetaInfos']
        req_json['type'] = 9
        req_json['cloudInfoEntity'] = {
            "cloudName": pool_name,
            "cloudDisplayName": pool_name
        }
        req_json['solrCloudPoolInfo'] = {
            'zookerEntity': {
                "_type_": 28,
                "clientGroupId": int(self._commcell_object.hac_clusters.get(hac_name).cluster_id),
                "clientGroupName": hac_name,
                "flags": {
                    "include": False
                }
            }
        }
        for node_name in node_names:
            node_obj = self._commcell_object.clients.get(node_name)
            node_data = {
                "opType": IndexServerConstants.OPERATION_ADD,
                "nodeClientEntity": {
                    "hostName": node_obj.client_hostname,
                    "clientId": int(node_obj.client_id),
                    "clientName": node_obj.client_name,
                    "_type_": 3
                },
                "nodeMetaInfos": [
                    {
                        "name": "ISENABLED",
                        "value": "false"
                    },
                    {
                        "name": "JVMMAXMEMORY",
                        "value": "8191"
                    },
                    {
                        "name": "PORTNO",
                        "value": "20000"
                    },
                    {
                        "name": "URL",
                        "value": ""
                    },
                    {
                        "name": "INDEXLOCATION",
                        "value": "NA"
                    }
                ]
            }
            req_json['cloudNodes'].append(node_data)
        flag, response = self._cvpysdk_object.make_request(
            'POST', self._services['CLOUD_CREATE'], req_json
        )
        if flag:
            if response.json() and 'genericResp' in response.json():
                if response.json()['genericResp'] == {} and \
                        'cloudId' in response.json():
                    self.refresh()
                    return IndexPool(self._commcell_object, pool_name)
                o_str = 'Failed to create index pool. Error: "{0}"'.format(
                    response.json()['genericResp'])
                raise SDKException('Response', '102', o_str)
            raise SDKException('Response', '102')
        self._response_not_success(response)

    def delete(self, pool_name):
        """Deletes an existing index pool cloud from commcell

        Args:
            pool_name       (str)   --  Index pool cloud name

        Returns:
            None

        Raises:
              SDKExecption:
                  Data type of the input(s) is not valid.

                  Response was not success.

                  Response was empty.
        """
        if not isinstance(pool_name, str):
            raise SDKException('IndexPools', '101')
        client = self.get(pool_name)
        cloud_id = client.cloud_id
        req_json = IndexServerConstants.REQUEST_JSON.copy()
        req_json["opType"] = IndexServerConstants.OPERATION_DELETE
        req_json['cloudInfoEntity']['cloudId'] = cloud_id
        flag, response = self._cvpysdk_object.make_request(
            'POST', self._services['CLOUD_DELETE'], req_json
        )
        if flag:
            if response.json() and 'genericResp' in response.json() \
                    and 'errorCode' not in response.json()['genericResp']:
                self.refresh()
                return
            if response.json() and 'genericResp' in response.json():
                raise SDKException(
                    'Response', '102', response.json()['genericResp'].get(
                        'errorMessage', ''))
            raise SDKException('Response', '102')
        self._response_not_success(response)


class IndexPool(object):
    """Class for performing index pool operations on a specific index pool"""

    def __init__(self, commcell_obj, pool_name, pool_id=None):
        """Initializes the IndexPool class instance

        Args:
            commcell_obj        (object)    --  Instance of class Commcell
            pool_name           (str)       --  Index pool name
            pool_id             (int)       --  Index pool client id
                default:
                    None

        Returns:
            object  -   instance of the IndexPool class

        """
        self.commcell = commcell_obj
        self._pool_name = pool_name
        self._pool_id = pool_id
        self.pool_client = None
        self.pool_properties = None
        self.pool_nodes = None
        self.refresh()

    def __repr__(self):
        """Representation string for the instance of the IndexPool class."""
        return "IndexPool class instance for Commcell: '{0}'".format(
            self.commcell.commserv_name
        )

    def _response_not_success(self, response):
        """Helper method to raise exception when response is not 200 (ok)

                Raises:
                    SDKException:
                        Response was not success
        """
        raise SDKException(
            'Response',
            '101',
            self.commcell._update_response_(
                response.text))

    def refresh(self):
        """Refreshes properties of IndexPool class"""
        self.commcell.clients.refresh()
        if not self.commcell.clients.has_client(self.pool_name):
            raise SDKException('IndexPools', '102')
        self.pool_client = self.commcell.clients.get(self.pool_name)
        self._pool_id = self.pool_client.client_id
        self.pool_properties = self.pool_client.\
            properties['pseudoClientInfo']['distributedClusterInstanceProperties']['clusterConfig']['cloudInfo']
        self.pool_nodes = self.pool_properties['cloudNodes']

    def node_info(self, node_name):
        """Returns the index pool node information

            Args:
                node_name       (str)       --  index pool node name

            Returns:
                dict    -   dictionary consisting details of index pool node

            Raises:
                SDKException:
                    Index pool node not found

        """
        for node_info in self.pool_nodes:
            if node_info['nodeClientEntity']['clientName'].lower() == node_name.lower():
                return node_info
        raise SDKException('IndexPools', '103')

    def modify_node(self, node_name,
                    operation_type=IndexServerConstants.OPERATION_EDIT,
                    node_params=None):
        """Method to modify the pool node details

        Args:
            node_name           (str)   --  index pool node name
            operation_type      (int)   --  operation type (1, 2, 3)
                                            1 - Adds a new node
                                            2 - Removes existing node
                                            3 - Edits the existing node (default)
            node_params         (list)  --  list of all the properties for the index pool node
                                            for example:
                                                [{
                                                    "name": <property_name>,
                                                    "value": <property_value>
                                                },
                                                ]
        Raises:
            SDKException:
                Response was not success.
                Response was empty.

        """
        req_json = {
            'cloudId': self.cloud_id,
            'type': 9,
            'nodes': [{
                'status': 1,
                'opType': operation_type,
                'nodeClientEntity': {
                    'clientId': int(self.commcell.clients[node_name]['id']),
                    'clientName': node_name
                },
                'nodeMetaInfos': []
            }]
        }
        if node_params:
            req_json['nodes'][0]['nodeMetaInfos'] = node_params
        flag, response = self.commcell._cvpysdk_object.make_request(
            'POST', self.commcell._services['CLOUD_NODE_UPDATE'],
            req_json
        )
        if flag:
            if response.json() is not None:
                if 'errorCode' not in response.json():
                    self.refresh()
                    return
        self._response_not_success(response)

    @property
    def pool_id(self):
        """Returns the index pool client id"""
        return self._pool_id

    @property
    def cloud_id(self):
        """Returns the index pool cloud id"""
        return self.pool_properties['cloudInfoEntity']['cloudId']

    @property
    def pool_name(self):
        """Returns index pool name"""
        return self._pool_name

    @property
    def node_names(self):
        """Returns a list of index pool node names"""
        result = []
        for node_info in self.pool_nodes:
            result.append(node_info['nodeClientEntity']['clientName'])
        return result

    @property
    def hac_cluster(self):
        """Returns the hac cluster name assigned to the index pool"""
        return self.pool_properties['solrCloudPoolInfo']['zookerEntity']['clientGroupName']



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

""" File for associating Workload and Backup destination regions for various entites
class: Regions. Region

Regions:
========
    _get_regions()                  --  Gets all the regions created in commcell
    refresh()                       --  Refresh the list of Regions associated to commcell
    has_region()                    --  Checks if region with given name exists
    get()                           --  returns Region class object for the specified input name
    set_region()                    --  Associate a region to an entity
    get_region()                    --  Gets the Region associated to an Entity
    calculate_region()              --  Calculates the Region to be associated to an Entity

Region:
=======
    _get_region_id()                --  Returns the region id

Attributes:
    ***region_id***                 --  Id of the given Region

"""
from .exception import SDKException


class Regions:
    """
    Class for representing all the Regions created in the commcell
    """
    def __init__(self, commcell_object):
        """Initialises the object of Regions class"""
        self._commcell_object = commcell_object
        self._cvpysdk_object = commcell_object._cvpysdk_object
        self._regions_api = self._commcell_object._services['REGIONS']
        self._EDIT_REGION = self._commcell_object._services['EDIT_REGION']
        self._GET_REGION = self._commcell_object._services['GET_REGION']
        self._CALCULATE_REGION = self._commcell_object._services['CALCULATE_REGION']
        self._regions = {}
        self.refresh()

    def _get_regions(self):
        """Gets all the regions created in commcell"""
        flag, response = self._cvpysdk_object.make_request('GET', self._regions_api)
        if flag:
            if response.json() and 'regions' in response.json():
                for region in response.json()['regions']:
                    name = region['name']
                    id = region['id']
                    self._regions[name] = id

                return self._regions

            raise SDKException('Response', '102')

        response_string = self._update_response_(response.text)
        raise SDKException('Response', '101', response_string)

    def refresh(self):
        """Refresh the list of Regions associated to commcell"""
        self._regions = self._get_regions()

    def has_region(self, name):
        """Checks if the given Region exists in the Commcell.

            Args:
                name    (str)   --  name of the Region

            Returns:
                bool    -   boolean output whether the Region exists in the commcell or not

            Raises:
                SDKException:
                    if type of the Region name argument is not string

        """
        if not isinstance(name, str):
            raise SDKException('Region', '103')

        return self._regions and (name in self._regions)

    def get(self, name):
        """
        Returns the instance of Region class for the given Region name
        Args:
            name    (str)   --  name of the Region

        Returns:
            object  -- Instance of Region class for the given Region name

        Raises:
            SDKException:
                - If the Region name argument is not a string

                - If No Region found in commcell with the given region name
        """
        if not isinstance(name, str):
            raise SDKException('Region', '102',"Invalid input received")

        if self.has_region(name):
            return Region(self._commcell_object, name, self._regions[name])
        raise SDKException('Region', '103',"Region not present in commcell")

    def set_region(self, entity_type, entity_id, entity_region_type, region_id):
        """
        Associate a region to an entity
        Args:
            entity_type         (str)   :   Type of the entity
                                            (eg:    COMMCELL,
                                                    COMPANY,
                                                    CLIENT,
                                                    CLIENT_GROUP,
                                                    MEDIAAGENT,
                                                    STORAGE_POOL, etc
                                            )
            entity_id           (int)   :   unique id of the entity

            entity_region_type  (str)   :   Type of the region
                                            (WORKLOAD or BACKUP)

            region_id           (int)   :   ID of the region from app_regions
        """
        request = {
            "entityRegionType": entity_region_type,
            "region":
                {
                    "id": region_id
                }
        }

        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'PUT', self._EDIT_REGION % (entity_type, entity_id), request
        )

        if flag:
            if response.json():
                error_code = response.json()['errorCode']

                if error_code != 0:
                    if error_code == 50000:
                        raise SDKException('Regions', '101')
                    elif error_code == 547:
                        raise SDKException('Regions', '102', 'Invalid regionID provided in request')
                    else:
                        error_string = response.json()['errorMessage']
                        raise SDKException('Regions', '102', '{0}'.format(error_string))

            else:
                raise SDKException('Response', '102')
        else:
            response_string = self._commcell_object._update_response_(response.text)
            raise SDKException('Response', '101', response_string)

    def get_region(self, entity_type, entity_id, entity_region_type):
        """
        Gets the Region associated to an Entity
        Args:
            entity_type         (str)   :   Type of the entity
                                            (eg:    COMMCELL,
                                                    COMPANY,


                                                    CLIENT,
                                                    CLIENT_GROUP,
                                                    MEDIAAGENT,
                                                    STORAGE_POOL, etc
                                            )
            entity_id           (int)   :   unique id of the entity

            entity_region_type  (str)   :   Type of the region
                                            (WORKLOAD or BACKUP)
        """
        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'GET', self._GET_REGION % (entity_type, entity_id, entity_region_type)
        )

        if flag:
            if response.json():
                try:
                    if response.json()['errorCode']:
                        error_string = response.json()['errorMessage']
                        raise SDKException('Regions', '102', '{0}'.format(error_string))

                except:
                    if response.json().get('regionId'):
                        return response.json().get('regionId')
                    return 0

            else:
                return None
        else:
            response_string = self._commcell_object._update_response_(response.text)
            raise SDKException('Response', '101', response_string)

    def calculate_region(self, entity_type, entity_id, entity_region_type):
        """
                Calculates the Region to be associated to an Entity
                Args:
                    entity_type         (str)   :   Type of the entity
                                                    (eg:    COMMCELL,
                                                            COMPANY,
                                                            CLIENT,
                                                            CLIENT_GROUP,
                                                            MEDIAAGENT,
                                                            STORAGE_POOL, etc
                                                    )
                    entity_id           (int)   :   unique id of the entity

                    entity_region_type  (str)   :   Type of the region
                                                    (WORKLOAD or BACKUP)
                """
        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'GET', self._CALCULATE_REGION % (entity_type, entity_id, entity_region_type)
        )

        if flag:
            if response.json():
                try:
                    if response.json()['errorCode']:
                        error_string = response.json()['errorMessage']
                        raise SDKException('Regions', '102', '{0}'.format(error_string))

                except:
                    return response.json()['regionId']

            else:
                raise SDKException('Response', '102')
        else:
            response_string = self._commcell_object._update_response_(response.text)
            raise SDKException('Response', '101', response_string)



class Region:
    """ Class for performing operations on a given Region """
    def __init__(self, commcell_object, region_name, region_id=None):
        """ Initialise the Region class instance.
            Args:
                commcell_object     (object)    --  instance of the Commcell class

                region_name         (str)       --  name of the region

                region_id           (int)       --  ID of the REgion
                                                    Default : None

            Returns:
                object  -   instance of the Region class
        """
        self._commcell_object = commcell_object
        self._region_name = region_name
        self._cvpysdk_object = commcell_object._cvpysdk_object
        self._update_response_ = commcell_object._update_response_

        self._region_name = region_name

        if region_id:
            self._region_id = str(region_id)
        else:
            self._region_id = self._get_region_id()

    def _get_region_id(self):
        """ Returns the ID of the Region """
        regions = Regions(self._commcell_object)
        id = regions.get(self._region_name).region_id

    @property
    def region_id(self):
        """ Get Region ID """
        return self._region_id


# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""File for operating on a Virtual Server Googlecloud Subclient.

GooglecloudVirtualServerSubclient is the only class defined in this file.

GooglecloudVirtualServerSubclient: Derived class from VirtualServerSubClient  Base class,
                                   representing a FusionCompute Subclient, and
                                   to perform operations on that Subclient

GooglecloudVirtualServerSubclient:

    __init__(,backupset_object, subclient_name, subclient_id)--  initialize object of googlecloud
                                                                             subclient object
                                                                                 associated with
                                                                        the VirtualServer subclient

    full_vm_restore_out_of_place()                  --  restores the VM  specified in
                                                     to the specified client, at the
                                                        specified destination location

    full_vm_restore_in_place()              --  restores the VM specified by the
                                                    user to the same location
"""

from ..vssubclient import VirtualServerSubclient
from ...exception import SDKException


class GooglecloudVirtualServerSubclient(VirtualServerSubclient):
    """Derived class from VirtualServerSubclient Base class.
       This represents a Google cloud virtual server subclient,
       and can perform restore operations on only that subclient.

    """
    def __init__(self, backupset_object, subclient_name, subclient_id=None):
        """Initialize the Instance object for the given Virtual Server instance.
        Args
        class_object (backupset_object, subclient_name, subclient_id)  --  instance of the
                                         backupset class, subclient name, subclient id

        """
        super(GooglecloudVirtualServerSubclient, self).__init__(
            backupset_object, subclient_name, subclient_id)
        self.diskExtension = ["none"]

    def full_vm_restore_in_place(
            self,
            vm_to_restore=None,
            overwrite=True,
            power_on=True,
            proxy_client=None,
            copy_precedence=0,
            zone=None,
            **kwargs):
        """Restores the FULL Virtual machine specified in the input list
            to the location same as the actual location of the VM in VCenter.

            Args:
                vm_to_restore       (list)     --  provide the list of VM name to
                                                   restore
                                                   default: None

                overwrite           (bool)     --  overwrite the existing VM
                                                   default: True

                power_on            (bool)     --  power on the  restored VM
                                                   default: True

                copy_precedence     (int)      --  copy precedence value
                                                   default: 0

                proxy_client          (str)  --  proxy client to be used for restore
                                                        default: proxy added in subclient

                **kwargs                         : Arbitrary keyword arguments Properties as of
                                                     full_vm_restore_in_place
                    eg:
                    v2_details          (dict)       -- details for v2 subclient
                                                    eg: check clients.vmclient.VMClient._child_job_subclient_details

            Returns:
                object - instance of the Job class for this restore job

            Raises:
                SDKException:
                    if inputs are not of correct type as per definition

                    if failed to initialize job

                    if response is empty

                    if response is not success

        """
        restore_option = {"v2_details": kwargs.get("v2_details", None), "datacenter": zone[:-2]}

        # set attr for all the option in restore xml from user inputs
        self._set_restore_inputs(
            restore_option,
            unconditional_overwrite=overwrite,
            power_on=power_on,
            copy_precedence=copy_precedence,
            vm_to_restore=self._set_vm_to_restore(vm_to_restore),
            volume_level_restore=1,
            client_name=proxy_client,
            in_place=True
        )

        request_json = self._prepare_fullvm_restore_json(restore_option)
        return self._process_restore_response(request_json)

    def full_vm_restore_out_of_place(
            self,
            vm_to_restore=None,
            proxy_client=None,
            new_name=None,
            zone=None,
            machine_type=None,
            overwrite=True,
            power_on=True,
            public_ip=False,
            copy_precedence=0,
            project_id=None,
            restore_option=None,
            **kwargs):

        """Restores the FULL Virtual machine specified  in the input  list to the client,
            at the specified destination location.


            Args:
                vm_to_restore     (list):       provide the list of VM name to restore
                                                default: None

                proxy_client     (str):         proxy client to be used for restore
                                                default: proxy added in subclient

                new_name         (str):         new name to be given to the
                                                restored VM

                overwrite         (bool):        overwrite the existing VM
                                                 default: True

                power_on          (bool):        power on the  restored VM
                                                  default: True

                copy_precedence   (int)        -- copy precedence value
                                                  default: 0

                **kwargs                         : Arbitrary keyword arguments Properties as of
                                                     full_vm_restore_out_of_place
                    eg:
                    v2_details          (dict)       -- details for v2 subclient
                                                    eg: check clients.vmclient.VMClient._child_job_subclient_details
                    destination_network (string)     -- Name of the destination network
                    networks_nic        (string)     -- Link for the destination network
                    subnetwork_nic      (string)    -- Link for the destination subnetwork

            Returns:
                object - instance of the Job class for this restore job

            Raises:
                SDKException:
                    if inputs are not of correct type as per definition

                    if failed to initialize job

                    if response is empty

                    if response is not success

        """
        restore_option = {}
        extra_options = ['destination_network', 'networks_nic', 'subnetwork_nic']
        for key in extra_options:
            if key in kwargs:
                restore_option[key] = kwargs[key]
            else:
                restore_option[key] = None

        if vm_to_restore:
            vm_to_restore = [vm_to_restore]

        if new_name:
            restore_option['restore_new_name'] = new_name

        if bool(restore_option):
            if not (isinstance(overwrite, bool) and
                    isinstance(power_on, bool)):
                raise SDKException('Subclient', '101')
        restore_option["datacenter"] = zone[:-2]
        if kwargs.get("replica_zone"):
            restore_option["replicaZones"] = []
            restore_option["replicaZones"].append(zone)
            restore_option["replicaZones"].append(kwargs.get("replica_zone"))

        # set attr for all the option in restore xml from user inputs
        self._set_restore_inputs(
            restore_option,
            vm_to_restore=self._set_vm_to_restore(vm_to_restore),
            unconditional_overwrite=overwrite,
            power_on=power_on,
            copy_precedence=copy_precedence,
            volume_level_restore=1,
            client_name=proxy_client,
            esx_host=zone,
            vm_size=machine_type,
            datacenter=zone,
            in_place=False,
            createPublicIP=public_ip,
            project_id=project_id,
            restore_new_name=new_name
        )

        request_json = self._prepare_fullvm_restore_json(restore_option)
        return self._process_restore_response(request_json)
